/* $Id: FlexArray.c 354 2004-07-17 06:28:14Z cher $ */

/* Copyright (C) 1995-2004 Alexander Chernov <cher@ispras.ru> */
/* Created: <95/12/03 14:54:18 cher> */

/*
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 */

#define __REUSE__ 1

/* internal include directives */
#include "lconfig.h"
#include "reuse/config.h"
#include "embed_version.h"

/* reuse include directives */
#include <reuse/integral.h>
#include <reuse/ReuseDefs.h>
#include <reuse/FlexArray.h>
#include <reuse/xalloc.h>

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>

#define FA_CLUSTER 32

typedef struct FA_INTERNAL
{
  fa_virtual_table *Virtual;
  size_t           Size;
  void             *Data;
  ruint_t             Used;
  ruint_t             Allocd;
  char             Dummy[1];
} FA_INTERNAL;

static void  fa_check(FA_INTERNAL *) REUSE_UNUSED;
static void  range_error(void) REUSE_UNUSED;

#ifdef REUSE_DEBUG
#define CHECK(p)         fa_check(p)
#define CHECKRANGE(fa,i) { if ((i) < 0 || ((unsigned int) i) > (fa)->Used) range_error(); }
#else
#define CHECK(p)
#define CHECKRANGE(fa,i)
#endif

  void
_faInsert(FA_INTERNAL* fa, int index)
{
  char *dest = NULL;

  CHECK(fa);
  CHECKRANGE(fa,index);

  if (fa->Used >= fa->Allocd)
    {
      fa->Allocd += FA_CLUSTER;
      fa->Data = xrealloc(fa->Data,fa->Allocd * fa->Size);
      assert(fa->Data != NULL);

      dest = (char*) fa->Data + index*fa->Size;
  
      if((unsigned int) index < fa->Used)
        {
          memmove (dest + fa->Size, dest, (fa->Used - index)*fa->Size);
        }
    }

  if(fa->Virtual->prc_copy != NULL)
    {
      (*fa->Virtual->prc_copy)(dest, &fa->Dummy);
    }
  else
    memcpy(dest, &fa->Dummy, fa->Size);

  fa->Used++;
}

  void
_faRemove(FA_INTERNAL *fa,int index)
{
  char *dest;
  
  CHECK(fa);
#ifdef REUSE_DEBUG
  if(index < 0 || (unsigned int) index >= fa->Used)
    {
      range_error();
    }
#endif

  dest = (char*) fa->Data + index*fa->Size;
  
  if(fa->Virtual->prc_destroy != NULL)
    (*fa->Virtual->prc_destroy)(dest);
  
  if((unsigned int) index < fa->Used - 1)
    {
      memmove (dest, dest+fa->Size, (fa->Used - 1 - index));
    }

  fa->Used--;
}

  void
_faClear(FA_INTERNAL *fa)
{
  CHECK(fa);

  if(fa->Virtual->prc_destroy != NULL)
    {
      int i;
      char *ptr;

      for(i = fa->Used, ptr = fa->Data; i ; i--, ptr += fa->Size)
        (*fa->Virtual->prc_destroy)((void*) ptr);
    }

  fa->Used = 0;
}

  void
_faDestroy(FA_INTERNAL *fa)
{
  CHECK(fa);

  if(fa->Virtual->prc_destroy != NULL)
    {
      int i;
      char *ptr;

      for(i = fa->Used, ptr = fa->Data; i ; i--, ptr += fa->Size)
        (*fa->Virtual->prc_destroy)((void*) ptr);
    }

  free(fa->Data);

  fa->Data = NULL;
  fa->Used = 0;
  fa->Allocd = 0;
}



  static void
fa_check(FA_INTERNAL *fa)
{
  if (fa->Virtual == NULL)
    {
      fputs("fa_check: virtual table is NULL!\n",stderr);
      abort();
    }

  if (fa->Size == 0)
    {
      fputs("fa_check: size is 0!\n",stderr);
      abort();
    }

  if (fa->Size >= 10000000)
    {
      fprintf(stderr, "fa_check: suspicious size %u!\n", fa->Size);
      abort();
    }

  if(fa->Used > fa->Allocd)
    {
      fputs("fa_check: fa->Used > fa->Allocd!\n",stderr);
      abort();
    }

  if(fa->Allocd > 0 && fa->Data == NULL)
    {
      fputs("fa_check: fa->Data == NULL when fa->Allocated > 0!\n",stderr);
      abort();
    }

  if(fa->Allocd == 0 && fa->Data != NULL)
    {
      fputs("fa_check: fa->Data != NULL when fa->Allocated == 0!\n",stderr);
      abort();
    }
}

  static void
range_error(void)
{
  fprintf(stderr,"Range checking error!\n");
  abort();
}

#if REUSE_COMPILE_CVSID - 0 != 0
static char const rcsid[] __attribute__((unused)) =
"$Id: FlexArray.c 354 2004-07-17 06:28:14Z cher $";
#endif /* REUSE_COMPILE_CVSID */

/**
 * Local variables:
 *  compile-command: "make"
 *  tags-file-name: "../TAGS"
 *  c-font-lock-extra-types: ("FILE" "\\sw+_t" "FA_INTERNAL" "fa_virtual_table")
 * End:
 */
