/* $Id: Memory.c 354 2004-07-17 06:28:14Z cher $ */

/* Copyright (C) 1999-2004 Alexander Chernov */
/* Ich, Doktor Josef Grosch, Informatiker, Sept. 1987 */

/*
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 */

#define __REUSE__ 1

/* internal include directives */
#include "lconfig.h"
#include "reuse/config.h"
#include "embed_version.h"

/* reuse include directives */
# include <reuse/Defs.h>
# include <reuse/Memory.h>
# include <reuse/System.h>
# include <reuse/General.h>

# include <stdio.h>

# define MinSizeSmallBlock      4
# define MaxSizeSmallBlock      62      /* 64 - 2       */
# define MinSizeLargeBlockLog   6       /* Log2 64      */
# define MaxSizeLargeBlockLog   24      /* Log2 2**24   */
# define PoolSize               10240L
# define NIL                    (tBlockPtr) NULL

static unsigned long MemoryUsed = 0;

struct tBlock {
   struct tBlock *      Successor;
   unsigned long        Size;
};
typedef struct tBlock * tBlockPtr;
typedef Cardinal        tSmallBlockRange;
typedef Cardinal        tLargeBlockRange;

static  tBlockPtr       SmallChain [MaxSizeSmallBlock    + 1] = { 0,
   NIL, NIL, NIL, NIL, NIL, NIL, NIL, NIL, NIL, NIL,
   NIL, NIL, NIL, NIL, NIL, NIL, NIL, NIL, NIL, NIL,
   NIL, NIL, NIL, NIL, NIL, NIL, NIL, NIL, NIL, NIL,
   NIL, NIL, NIL, NIL, NIL, NIL, NIL, NIL, NIL, NIL,
   NIL, NIL, NIL, NIL, NIL, NIL, NIL, NIL, NIL, NIL,
   NIL, NIL, NIL, NIL, NIL, NIL, NIL, NIL, NIL, NIL,
   NIL, NIL,
};
static  tBlockPtr       LargeChain [MaxSizeLargeBlockLog + 1] = { 0,
   NIL, NIL, NIL, NIL, NIL, NIL, NIL, NIL, NIL, NIL,
   NIL, NIL, NIL, NIL, NIL, NIL, NIL, NIL, NIL, NIL,
   NIL, NIL, NIL, NIL,
};
static  char *          PoolFreePtr = 0;
static  char *          PoolEndPtr  = 0;

static short initialized;
  static void
do_init()
{
  register int i;

  initialized = 1;

  for (i = MinSizeSmallBlock; i <= MaxSizeSmallBlock; i += 2) {
    SmallChain [i] = NIL;
  }

  for (i = MinSizeLargeBlockLog; i <= MaxSizeLargeBlockLog; i ++) {
    LargeChain [i] = NIL;
  }

  MemoryUsed   = 0;
  PoolFreePtr  = 0;
  PoolEndPtr   = 0;
}

/**
 * Returns a pointer to dynamically allocated
 * space of size 'ByteCount' bytes.
 */
  char *
Alloc(unsigned long ByteCount)
{
  if (!initialized) do_init();

  ByteCount = (ByteCount + yyMaxAlign - 1) & yyAlignMasks [yyMaxAlign];

  if (ByteCount <= MaxSizeSmallBlock) {
     /* handle small block */

    if (ByteCount < MinSizeSmallBlock) ByteCount = MinSizeSmallBlock;
    if (SmallChain [ByteCount] != NIL) {
      /* obtain block from freelist */
      register tBlockPtr CurrentBlock = SmallChain [ByteCount];
      SmallChain [ByteCount] = CurrentBlock->Successor;
      return (char *) CurrentBlock;
    } else {
      /* obtain block from storage pool */
      register unsigned long FreeBytes;
      register char *        OldFreePtr;

      /* release old pool */
      if ((FreeBytes = PoolEndPtr - PoolFreePtr) < ByteCount) {
        if (FreeBytes >= MinSizeSmallBlock) Free (FreeBytes, PoolFreePtr);
        PoolFreePtr = Alloc (PoolSize);     /* allocate new pool */
        PoolEndPtr  = PoolFreePtr + PoolSize;
      }
      OldFreePtr = PoolFreePtr;
      PoolFreePtr += ByteCount;
      return OldFreePtr;
    }
  } else {
    /* handle large block */

    /* 1. search in LargeChain [Log2 (ByteCount)] using BEST FIT */

    register Cardinal         ChainNumber     = Log2 (ByteCount);
    register tBlockPtr        CurrentBlock    = LargeChain [ChainNumber];
    register tBlockPtr        PreviousBlock   = (tBlockPtr) & (LargeChain [ChainNumber]);
    register tBlockPtr        BestBlock       = NIL;
    register unsigned long    BestBlockSize   = 1000000000;
    register tBlockPtr        PredecessorBlock = NULL;
    register tLargeBlockRange j               ;
    register unsigned long    CurrentBlockSize;

    while (CurrentBlock) {
      CurrentBlockSize = CurrentBlock->Size;
      if (CurrentBlockSize >= ByteCount) {
        if (CurrentBlockSize == ByteCount) {
          /* exact match */
          PreviousBlock->Successor = CurrentBlock->Successor;
          return (char *) CurrentBlock;
        }
        if (CurrentBlockSize < BestBlockSize) {
          /* improve approximation */
          BestBlock        = CurrentBlock;
          BestBlockSize    = CurrentBlockSize;
          PredecessorBlock = PreviousBlock;
        }
      }
      PreviousBlock  = CurrentBlock;
      CurrentBlock   = CurrentBlock->Successor;
    }

    if (BestBlock) {
      PredecessorBlock->Successor = BestBlock->Successor;
      if (BestBlockSize - ByteCount >= MinSizeSmallBlock) {
        Free (BestBlockSize - ByteCount, (char *) BestBlock + ByteCount);
      }
      return (char *) BestBlock;
    }

    /* 2. search in LargeChain [j], j > Log2 (ByteCount), using FIRST FIT */
    for (j = ChainNumber+1; j <= MaxSizeLargeBlockLog; j ++) {
      CurrentBlock = LargeChain [j];
      if (CurrentBlock != NIL) {
        LargeChain [j] = CurrentBlock->Successor;
        if (CurrentBlock->Size - ByteCount >= MinSizeSmallBlock) {
          Free (CurrentBlock->Size - ByteCount,
                (char *) CurrentBlock + ByteCount);
        }
        return (char *) CurrentBlock;
      }
    }

    if (ByteCount < PoolSize) {
      /* 3. obtain block from storage pool */
      register unsigned long FreeBytes;

      /* release old pool */
      if ((FreeBytes = PoolEndPtr - PoolFreePtr) < ByteCount) {
        if (FreeBytes >= MinSizeSmallBlock) Free (FreeBytes, PoolFreePtr);
        PoolFreePtr = Alloc (PoolSize);     /* allocate new pool */
        PoolEndPtr  = PoolFreePtr + PoolSize;
      }
      PoolFreePtr += ByteCount;
      return PoolFreePtr - ByteCount;
    } else {
      /* 4. allocate individual block */
      CurrentBlock = (tBlockPtr) SysAlloc ((long) ByteCount);
      MemoryUsed += ByteCount;
      return (char *) CurrentBlock;
    }
  }
}

/**
 * The dynamically allocated space starting at
 * address 'a' of size 'ByteCount' bytes is
 * released.
 */
  void
Free (unsigned long ByteCount, char *a)
{
   register tBlockPtr           BlockPtr;
   register tLargeBlockRange    ChainNumber;

   if (!initialized) do_init();

   ByteCount = (ByteCount + yyMaxAlign - 1) & yyAlignMasks [yyMaxAlign];

   BlockPtr = (tBlockPtr) a;
   if (ByteCount <= MaxSizeSmallBlock) {
      if (ByteCount < MinSizeSmallBlock) ByteCount = MinSizeSmallBlock;
      BlockPtr->Successor       = SmallChain [ByteCount];
      SmallChain [ByteCount]    = BlockPtr;
   } else {
      ChainNumber               = Log2 (ByteCount);
      BlockPtr->Successor       = LargeChain [ChainNumber];
      BlockPtr->Size            = ByteCount;
      LargeChain [ChainNumber]  = BlockPtr;
   }
}

  void
InitMemory (void)
{
  if (!initialized) do_init();
}

#if REUSE_COMPILE_CVSID - 0 != 0
static char const rcsid [] =
"$Id: Memory.c 354 2004-07-17 06:28:14Z cher $";
#endif /* REUSE_COMPILE_CVSID */

/**
 * Local variables:
 *  compile-command: "make"
 *  tags-file-name: "TAGS"
 *  c-font-lock-extra-types: ("\\sw+_t" "FILE" "Cardinal" "tBlockPtr" "tLargeBlockRange")
 * End:
 */
