/* $Id: RErrors.c 354 2004-07-17 06:28:14Z cher $ */

/* Copyright (C) 1995-2004 Alexander Chernov <cher@ispras.ru> */
/* Ich, Doktor Josef Grosch, Informatiker, Juli 1992 */
/* Alexander Chernov, Nov. 1995 */

/*
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 */

#define __REUSE__ 1

/* internal include directives */
#include "lconfig.h"
#include "reuse/config.h"
#include "embed_version.h"

/* reuse include directives */
#include <reuse/integral.h>
#include <reuse/RErrors.h>
#include <reuse/Sets.h>
#include <reuse/Idents.h>
#include <reuse/GenOut.h>
#include <reuse/FlexString.h>
#include <reuse/Printf.h>

#include <assert.h>
#include <string.h>

#define CLUSTER             16

typedef struct tErrorStore
{
  tPosition Pos;
  int       Class;
  int       Number;
  char      *Message;
} tErrorStore;

typedef void (*tErrorHandler)(int,int,tPosition,char*);
typedef int  (*tCompareFunc)(const void *,const void *);

#define strTempBuf_len     2048

static tErrorStore *persStoredErrors;
static ruint_t        ulStoreSize;
static ruint_t        ulStoreUsed;
static int         *piErrorCodeHash;
static int         iMaxErrorNumber;
static int         fStoreMsg;
static char        strTempBuf[strTempBuf_len];
static tErrorTable *petErrors;
/*static tFileOut    goFileOutput;*/
static tFixedOut   goFixedOutput;
static FILE        *errOutput = 0;

static tErrorHandler pehHandler;

static int errOccurences[xxErrorClasses];
static void (*errFatalHook)(void) = NULL;

static mf_type mfFormat;

static int errDisableTable[xxErrorClasses];

static void errPrintError(int, int, tPosition, char *);

static char *errClassesTable[]=
{
  "",
  "FATAL:",
  "Restriction:",
  "Error:",
  "Warning:",
  "Repair:",
  "Note:",
  "Information:",
  "FATAL INTERNAL:"
};

static char *strAltClassesTable[]=
{
  "",
  "FATAL ",
  "RESTRICTION ",
  "ERROR ",
  "WARNING ",
  "REPAIR ",
  "NOTE ",
  "INFORMATION ",
  "FATAL INTERNAL "
};


static Bool initialized;
static void do_init2(void)
{
  int i;

  initialized = True;

/* clear error store */
  persStoredErrors = NULL;
  ulStoreSize = 0;
  ulStoreUsed = 0;

/* clear occurences */
  for(i=0;i<xxErrorClasses;errOccurences[i++] = 0);
  fStoreMsg = 0;
  petErrors = NULL;
  pehHandler = errPrintError;
  
  /* set position printing format */
  mfFormat = mf_standard;
  errOutput = stderr;
}

static void do_init(void)
{
  do_init2();
  posInitModule();
}

  static void
WriteCode(int ErrorCode,tGenOut *pOut)
{
  if (!initialized) do_init();

  switch (ErrorCode)
    {
    case xxNoText:
      break;
    case xxSyntaxError:
      go_puts(pOut,"syntax error ");
      break;
    case xxExpectedTokens:
      go_puts(pOut,"expected tokens ");
      break;
    case xxRestartPoint:
      go_puts(pOut,"restart point ");
      break;
    case xxTokenInserted:
      go_puts(pOut,"token inserted ");
      break;
    case xxTooManyErrors:
      go_puts(pOut,"too many errors ");
      break;
    default:
      go_printf (pOut," error code: %d", ErrorCode);
    }
}

  static void
WriteInfo(int InfoClass,tGenOut *pOut,void *Info)
{
   int i;

   if (!initialized) do_init();
   
   if (InfoClass == xxNone) return;
   
   switch (InfoClass)
     {
     case xxInteger:
       go_printf(pOut,"%d",*(int*) Info);
       break;
     case xxShort:
       i =  * (short *) Info;
       go_printf(pOut,"%d", i);
       break;
     case xxLong:
       go_printf(pOut,"%ld",*(long *) Info);
       break;
     case xxReal:
       go_printf(pOut,"%e",*(float *) Info);
       break;
     case xxBoolean:
       go_puts(pOut,(*(Bool*) Info)?"TRUE":"FALSE");
       break;
     case xxCharacter:
       go_putc(pOut,*(char*) Info);
       break;
     case xxString:
       go_puts(pOut,(char*) Info);
       break;
     case xxSet:
       setWrite(pOut,(tSet *) Info,"{"," ","}");
       break;
     case xxIdent:
       go_puts(pOut,idGetString(* (tIdent*) Info));
       break;
     default:
       go_printf(pOut,"info class: %d", InfoClass);
       break;
   }
   
   go_putc(pOut,':');
}

  static int
errCmpErrorStore(const tErrorStore *p1,const tErrorStore *p2)
{
  return posCmp(&p1->Pos,&p2->Pos);
}

  static void
errAddError(int iClass,int iNumber,tPosition posPos,char *strMessage)
{
  assert(ulStoreSize >= ulStoreUsed);
  if (!initialized) do_init();

  if(ulStoreUsed == ulStoreSize)
    {
      persStoredErrors =
        (tErrorStore*) realloc(persStoredErrors,
                               sizeof(tErrorStore)*(ulStoreSize += CLUSTER));
      assert(persStoredErrors != NULL);
    }
  persStoredErrors[ulStoreUsed].Pos = posPos;
  persStoredErrors[ulStoreUsed].Class = iClass;
  persStoredErrors[ulStoreUsed].Number = iNumber;
  persStoredErrors[ulStoreUsed].Message = strdup(strMessage);
  assert(persStoredErrors[ulStoreUsed].Message != NULL);
  ulStoreUsed ++;
  
  if(iClass == xxFatal)
    {
      errFlushMessages(stderr);
      if(errFatalHook != NULL)
        (*errFatalHook)();
      exit(2);
    }
  
  if(iClass == xxAbort)
    {
      errFlushMessages(stderr);
      abort();
    }
}

  static void
errPrintError(int iClass,int iNumber,tPosition pos,char *strMessage)
{
  if (!initialized) do_init();

  if(pos.Line != 0 || pos.Column != 0 || pos.FName != 0)
    {
      switch(mfFormat)
        {
        case mf_standard:
          posPrintf(errOutput,"%f:%l:%c:",pos);
          break;
        case mf_no_file:
          posPrintf(errOutput,"%l:%c:",pos);
          break;
        case mf_SDT:
          posPrintf(errOutput,"%g\n",pos);
          break;
        }
    }

  if(iClass < 0 || iClass >= xxErrorClasses)
    {
      fprintf(errOutput,"Unknown Error Class (%d):",iClass);
    }
  else
    {
      if(mfFormat == mf_SDT)
        fputs(strAltClassesTable[iClass],errOutput);
      else
        fputs(errClassesTable[iClass],errOutput);
    }

  if(mfFormat == mf_SDT)
    {
      if(iNumber == erUnknown)
        fputs("(UNKNOWN) ",errOutput);
      else
        fprintf(errOutput,"%d ",iNumber);
    }
  
  fputs(strMessage,errOutput);
  putc('\n',errOutput);

  if(iClass == xxFatal)
    {
      if(errFatalHook != NULL)
        (*errFatalHook)();
      exit(2);
    }
  
  if(iClass == xxAbort)
    {
      errFlushMessages(stderr);
      abort();
    }
  
  if(iClass >=0 && iClass < xxErrorClasses)
    errOccurences[iClass]++;
}

  void
errSetMessageFormat(int mfNewFormat)
{
  if (!initialized) do_init();

  mfFormat = mfNewFormat;
}

  void
_errInitModule(int sysmode, tErrorTable *etTable, int mfNewFormat)
{
  int i;

  assert(sysmode >= 0 && sysmode <= 1);

  if(initialized) return;

  do_init2();
  posInitModule();

  iMaxErrorNumber = 0;  
  if (etTable != NULL)
    {
/* count number of messages */
      for(i=0;etTable[i].String != NULL;i++)
        {
          if(etTable[i].Number < 0)
            {
              fputs("Cannot handle errors with numbers less than 0\n",
                    stderr);
              abort();
            }
          if(etTable[i].Number > iMaxErrorNumber)
            iMaxErrorNumber = etTable[i].Number;
        }
      
/* allocate space for hash */
      piErrorCodeHash = (int*) malloc(sizeof(int)*(iMaxErrorNumber+1));
      assert(piErrorCodeHash != NULL);
      
/* fill it with -1 */
      for(i=0;i<=iMaxErrorNumber;piErrorCodeHash[i++] = -1);
      
/* fill the hash */
      for(i=0;etTable[i].String != NULL;i++)
        {
          int j = etTable[i].Number;
          
          if(piErrorCodeHash[j] != -1)
            {
              fputs("Cannot handle duplicated error numbers\n",stderr);
              abort();
            }
          piErrorCodeHash[j] = i;
        }
    }
}

void
errCloseModule(void)
{
  initialized = False;
}

  void
errVMessage(int iClass,tPosition pos,char *strFormat,va_list vaArgs)
{
  tFixedOut fixed_out;
  
  if (!initialized) do_init();

  fixed_out = goFixedOut (strTempBuf, strTempBuf_len);
  
  VGPrintf ((tGenOut*) &fixed_out, strFormat,vaArgs);
  
  (*pehHandler)(iClass,erUnknown,pos,strTempBuf);
}

  void
errMessage(int iClass,tPosition pos,char *strFormat,...)
{
  va_list vaArgs;
  
  if (!initialized) do_init();

  va_start(vaArgs,strFormat);
  errVMessage(iClass,pos,strFormat,vaArgs);
  va_end(vaArgs);
}

  void
errVReport(int iCode,tPosition pos,va_list vaArgs)
{
  char *strErrorMsg;
  tFixedOut fixed_out;

  if (!initialized) do_init();

  if(iCode < 0 || iCode > iMaxErrorNumber || piErrorCodeHash[iCode] == -1)
    strErrorMsg = "Unknown error";
  else
    strErrorMsg = petErrors[piErrorCodeHash[iCode]].String;
  
  fixed_out = goFixedOut (strTempBuf, strTempBuf_len);
  
  VGPrintf((tGenOut*) &fixed_out, strErrorMsg,vaArgs);
  
  (*pehHandler)(petErrors[piErrorCodeHash[iCode]].Class,iCode,pos,strTempBuf);
}

  void
errReport(int iCode,tPosition pos,...)
{
  va_list vaArgs;
  
  if (!initialized) do_init();

  va_start(vaArgs,pos);
  errVReport(iCode,pos,vaArgs);
  va_end(vaArgs);
}

  void
errFlushMessages(FILE *f)
{
  unsigned int i;
  
  if (!initialized) do_init();

  if(f == NULL)
    f = stderr;
  errOutput = f;
  
  if(ulStoreUsed == 0)
    return;
  qsort(persStoredErrors,ulStoreUsed,sizeof(tErrorStore),
        (tCompareFunc) errCmpErrorStore);
  
  for(i=0;i<ulStoreUsed;i++)
    errPrintError(persStoredErrors[i].Class,
                  persStoredErrors[i].Number,
                  persStoredErrors[i].Pos,
                  persStoredErrors[i].Message);
    
  ulStoreUsed = 0;
  errOutput = stderr;
}

  void
errStoreMessages(int fStore)
{
  if (!initialized) do_init();

  if(fStoreMsg && !fStore)
    {
      ulStoreUsed = 0;
    }
  fStoreMsg = !!(fStore);
  
  if(fStoreMsg)
    pehHandler = errAddError;
  else
    pehHandler = errPrintError;
}

  void
ErrorMessage(int ErrorCode, int ErrorClass, tPosition Position)
{
  tGenOut *pOut;
  
  if (!initialized) do_init();

  if(errDisableTable[ErrorCode])
    return;
  
  goFixedOutput = goFixedOut(strTempBuf,strTempBuf_len);
  pOut = (tGenOut*) &goFixedOutput;
  
  WriteCode(ErrorCode,pOut);
  (*pehHandler)(ErrorClass,ErrorCode,Position,strTempBuf);
}

  void
ErrorMessageI(int ErrorCode, int ErrorClass, tPosition Position,
              int InfoClass, void *Info)
{
  tGenOut *pOut;
  
  if (!initialized) do_init();

  if(errDisableTable[ErrorCode])
    return;
  
  goFixedOutput = goFixedOut(strTempBuf,strTempBuf_len);
  pOut = (tGenOut*) &goFixedOutput;
  
  WriteCode(ErrorCode,pOut);
  WriteInfo(InfoClass,pOut,Info);
  (*pehHandler)(ErrorClass,ErrorCode,Position,strTempBuf);
}

  void
Message(char * ErrorText, int ErrorClass, tPosition Position)
{
  if (!initialized) do_init();

  (*pehHandler)(ErrorClass,erUnknown,Position,ErrorText);
}

  void
MessageI(char * ErrorText,int ErrorClass,
         tPosition Position,int InfoClass,void *Info)
{
  tGenOut *pOut;
  
  if (!initialized) do_init();

  goFixedOutput = goFixedOut(strTempBuf,strTempBuf_len);
  pOut = (tGenOut*) &goFixedOutput;
  
  go_puts(pOut,ErrorText);
  go_putc(pOut, ' ');
  WriteInfo(InfoClass,pOut,Info);
  
  (*pehHandler)(ErrorClass,erUnknown,Position,strTempBuf);
}

void
errDisable(int ErrorCode)
{
  if (!initialized) do_init();

  if(ErrorCode > 0 && ErrorCode < xxErrorCodes)
    errDisableTable[ErrorCode] = 1;
}

int
errStatistics(FILE *f,int iMode)
{
  int summ = 0;
  int rcode = 0;
  
  if(iMode == 1)
    {
      if(errOccurences[xxError] > 0)
        return 1;
      else return 0;
    }
  
  fprintf(f,"Compilation statistics:\n");
  if(errOccurences[xxError] > 0)
    {
      fprintf(f,"Errors:         %d\n",errOccurences[xxError]);
      rcode = 1;
      summ += errOccurences[xxError];
    }
  
  if(errOccurences[xxWarning] > 0)
    {
      fprintf(f,"Warnings:       %d\n",errOccurences[xxWarning]);
      summ += errOccurences[xxWarning];
    }
  
  if(errOccurences[xxRestriction] > 0)
    {
      fprintf(f,"Restrictions:   %d\n",errOccurences[xxRestriction]);
      summ += errOccurences[xxRestriction];
    }
  
  if(errOccurences[xxRepair] > 0)
    {
      fprintf(f,"Repairs:        %d\n",errOccurences[xxRepair]);
      summ += errOccurences[xxRepair];
    }
  
  if(errOccurences[xxNote] > 0)
    {
      fprintf(f,"Notes           %d\n",errOccurences[xxNote]);
      summ += errOccurences[xxNote];
    }
  
  if(errOccurences[xxInformation] > 0)
    {
      fprintf(f,"Informations:   %d\n",errOccurences[xxInformation]);
      summ += errOccurences[xxInformation];
    }
  
  if(summ > 0)
    {
      fprintf(f,"Total messages: %d\n",summ);
    }
  else
    {
      fprintf(f,"No errors or warnings or other messages.\n");
    }
  
  return rcode;
}

  void
errClear(void)
{
  if (!initialized) do_init();
  memset(errOccurences, 0, sizeof(errOccurences));
}

  int *
errGetMsgCounters(void)
{
  if (!initialized) do_init();
  return errOccurences;
}

  errFatalHookFunc_t
errGetFatalHook(void)
{
  if (!initialized) do_init();
  return errFatalHook;
}

  errFatalHookFunc_t
errSetFatalHook(errFatalHookFunc_t new)
{
  errFatalHookFunc_t old;

  if (!initialized) do_init();
  old = errFatalHook;
  errFatalHook = new;
  return old;
}

#if REUSE_COMPILE_CVSID - 0 != 0
static char const rcsid [] =
"$Id: RErrors.c 354 2004-07-17 06:28:14Z cher $";
#endif /* REUSE_COMPILE_CVSID */

/**
 * Local variables:
 *  compile-command: "make"
 *  c-font-lock-extra-types: ("\\sw+_t" "FILE" "tGenOut" "tPosition" "va_list" "tErrorStore" "tCompareFunc" "tFixedOut" "tFileOut" "tErrorTable" "tErrorHandler" "mf_type" "Bool")
 *  tags-file-name: "TAGS"
 * End:
 */
