/* $Id: exec.c 57 2001-06-23 07:04:26Z cher $ */

/* Copyright (C) 2000-2001 Alexander Chernov <cher@ispras.ru> */
/* Created: <2000-07-01 20:07:32 cher> */

/*
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 */

/**
 * FILE:    utils/exec.c
 * PURPOSE: process abstraction layer
 */

static char const cvsid[] =
"$Id: exec.c 57 2001-06-23 07:04:26Z cher $";

#include "lconfig.h"
#include <p_config.h>

#include CONF_DEPFILE("reuse_version.h")

#include "reuse/exec.h"
#include "reuse/xalloc.h"
#include "reuse/logger.h"
#include "reuse/osdeps.h"

#include <stdio.h>
#include <stdarg.h>
#include <errno.h>

#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>

#include <process.h>

/* file descriptor redirections */
typedef struct tRedir
{
  int       fd;                 /* file descriptor */
  int       tag;                /* file descriptor operation */
  union
  {
    int       fd2;              /* second fd for dup operation */
    struct
    {
      char     *path;           /* file path */
      int       oflag;          /* flags for open(2) call */
      int       mode;           /* mode for open(2) call */
    }         s;                /* file redirection structure */
  }         u;
} tRedir;

/* task description structure */
struct tTask
{
  int  (*main)(int, char **);   /* task start function */
  char  *path;                  /* task invocation path */
  int    state;                 /* current task state */
  int    code;                  /* process termination code */
  int    is_exited;             /* is process exited normally */
  int    exit_code;             /* process exit code */
  struct
  {
    char **v;                   /* pointer to data */
    int    u;                   /* used entries */
    int    a;                   /* allocated entries */
  }      args;                  /* arguments to pass to subprocess */
  struct
  {
    tRedir *v;                  /* pointer to data */
    int     u;                  /* used entries */
    int     a;                  /* allocated entries */
  } redirs;                     /* redirections */
};

#define REUSE_TERMSIG(x)    ((unsigned int) (x) >> 8)
#define REUSE_EXITSTATUS(x) ((x) & 0xFF)

static tTask *cur_task = 0;     /* no multitasking under MSDOS */
static int    verbose_flag;

/**
 * NAME:    task_SetFlag
 * PURPOSE: set module options
 * ARGS:    opt  - option name
 *          flag - option id
 * RETURN:  0
 * NOTE:    this function is called back from getopt module
 */
  int
task_SetFlag(char *opt, int flag)
{
  switch (flag)
    {
    case 200:
      verbose_flag = 1;
      break;
    case 201:
      verbose_flag = 0;
      break;
    default:
      SWERR(("task_SetFlags: unsupported flag = %d", flag));
    }
  return 0;
}

/**
 * NAME:    task_New
 * PURPOSE: create new task structure
 * RETURN:  pointer to newly created task structure
 */
  tTask *
task_New(void)
{
  if (cur_task != NULL) return NULL;

  XCALLOC(cur_task, 1);
  cur_task->state = TSK_STOPPED;
  return cur_task;
}

/**
 * NAME:    task_Delete
 * PURPOSE: delete task structure
 * ARGS:    tsk - task to delete
 * NOTE:    if task is active when deleted, task is not terminated
 *          but detached and not watched for exit
 */
  void
task_Delete(tTask *tsk)
{
  int i;

  if (!tsk) return;
  ASSERT(tsk == cur_task);

  for (i = 0; i < tsk->args.u; i++)
    xfree(tsk->args.v[i]);
  for (i = 0; i < tsk->redirs.u; i++) {
    if (tsk->redirs.v[i].tag == TSR_FILE) {
      xfree(tsk->redirs.v[i].u.s.path);
    }
  }
  xfree(tsk->path);
  xfree(tsk->args.v);
  xfree(tsk->redirs.v);
  xfree(tsk);
  cur_task = NULL;
}

/**
 * NAME:    task_GetPipe
 * PURPOSE: get the actual fd of pipe redirections
 * ARGS:    tsk - task structure
 *          fd  - piped file descriptor
 * RETURNS: >= 0 - fd of the pipe
 *           < 0 - error
 */
  int
task_GetPipe(tTask *tsk, int fd)
{
  ASSERT(tsk != NULL);
  ASSERT(tsk == cur_task);
  return -1;
}

/**
 * NAME:    do_add_arg
 * PURPOSE: add argument to the task start arguments
 * ARGS:    tsk - task to add argument to
 *          arg - argument to add
 */
  static void
do_add_arg(tTask *tsk, char const *arg)
{
  if (tsk->args.u + 1 >= tsk->args.a)
    {
      tsk->args.a += 16;
      tsk->args.v = (char**) xrealloc(tsk->args.v,
                                      sizeof(char*) * tsk->args.a);
    }
  tsk->args.v[(tsk->args.u)++] = xstrdup(arg);
  tsk->args.v[tsk->args.u] = NULL;
}

/**
 * NAME:    task_AddArg
 * PURPOSE: add argument to task start arguments
 * ARGS:    tsk - task
 *          arg - argument
 */
  int
task_AddArg(tTask *tsk, char const *arg)
{
  ASSERT(tsk);
  ASSERT(tsk == cur_task);
  if (!arg) arg = "";

  if (tsk->state != TSK_STOPPED)
    {
      return -1;
    }

  do_add_arg(tsk, arg);
  return 0;
}

/**
 * NAME:    task_nAddArgs
 * PURPOSE: add arguments to the tast invokation arguments
 * ARGS:    tsk - task
 *          n   - number of arguments
 *          ... - arguments (char* pointers)
 * RETURN:  0 - ok, -1 - error
 */
  int
task_nAddArgs(tTask *tsk, int n, ...)
{
  va_list args;

  ASSERT(tsk);
  ASSERT(tsk == cur_task);

  if (tsk->state != TSK_STOPPED)
    {
      return -1;
    }

  va_start(args, n);
  for (; n > 0; n--)
    do_add_arg(tsk, va_arg(args, char *));
  va_end(args);

  return 0;
}

/**
 * NAME:    task_zAddArgs
 * PURPOSE: add task invokation arguments
 * ARGS:    tsk - task
 *          ... - extra arguments (terminated with NULL argument)
 * RETURN:  0 - ok, -1 - error
 */
  int
task_zAddArgs(tTask *tsk, ...)
{
  va_list  args;
  char    *s;

  ASSERT(tsk);
  ASSERT(tsk == cur_task);

  if (tsk->state != TSK_STOPPED)
    {
      return -1;
    }

  va_start(args, tsk);
  while ((s = va_arg(args, char *)))
    do_add_arg(tsk, s);
  va_end(args);

  return 0;
}

/**
 * NAME:    task_pnAddArgs
 * PURPOSE: add arguments to task invokation parameters
 * ARGS:    tsk - task
 *          n   - number of arguments
 *          p   - array of arguments
 * RETURN:  0 - ok, -1 - error
 */
  int
task_pnAddArgs(tTask *tsk, int n, char **p)
{
  ASSERT(tsk);
  ASSERT(tsk == cur_task);

  if (tsk->state != TSK_STOPPED)
    {
      return -1;
    }

  if (!p) return 0;
  for (; n > 0; n--)
    do_add_arg(tsk, *p++);

  return 0;
}

/**
 * NAME:    task_pzAddArgs
 * PURPOSE: add arguments to task invokation arguments
 * ARGS:    tsk - task
 *          p   - array of arguments (terminated with NULL)
 * RETURN:  0 - ok, -1 - error
 */
  int
task_pzAddArgs(tTask *tsk, char **p)
{
  ASSERT(tsk);
  ASSERT(tsk == cur_task);

  if (tsk->state != TSK_STOPPED)
    {
      return -1;
    }

  if (!p) return 0;
  while (*p)
    do_add_arg(tsk, *p++);

  return 0;
}

/**
 * NAME:    task_SetPath
 * PURPOSE: set invokation path
 * ARGS:    tsk - task
 *          arg - invocation path
 * RETURN:  0 - ok, -1 - error
 */
  int
task_SetPath(tTask *tsk, char const *arg)
{
  ASSERT(tsk);
  ASSERT(tsk == cur_task);
  ASSERT(arg);

  if (tsk->state != TSK_STOPPED)
    return -1;

  xfree(tsk->path);
  tsk->path = xstrdup(arg);
  return 0;
}

/**
 * NAME:    task_SetPathAsArg0
 * PURPOSE: use argv[0] as invokation path
 * ARGS:    tsk - task
 * RETURN:  invocation path, or NULL if error
 * NOTE:    the function is deprecated, argv[0] is used by default
 *          if explicit invocation path is not set
 */
  char *
task_SetPathAsArg0(tTask *tsk)
{
  ASSERT(tsk);
  ASSERT(tsk == cur_task);

  if (tsk->state != TSK_STOPPED)
    return NULL;

  if (tsk->args.u <= 0)
    return NULL;
  xfree(tsk->path);
  tsk->path = xstrdup(tsk->args.v[0]);
  return tsk->path;
}

/**
 * NAME:    task_SetRedir
 * PURPOSE: set file descriptiors redirections
 * ARGS:    tsk  - task
 *          fd   - file descriptor
 *          mode - redirection mode
 *          ...  - redirection mode specific arguments
 * RETURN:  0 - ok, -1 - error
 * NOTE:    redirection mode specific arguments are as follows:
 *            for TSR_CLOSE:
 *              no specific arguments
 *            for TSR_FILE
 *              char *path  - file path
 *              int   oflag - open flags
 *              int   mode  - open file mode
 *            for TSR_DUP
 *              int fd2     - second file descriptor
 *            for TSR_PIPE
 *              int   idx   - inherited index of the pipe
 *                            (0 - read, 1 - write)
 */
  int
task_SetRedir(tTask *tsk, int fd, int mode, ...)
{
  int     i;
  va_list args;

  ASSERT(tsk);
  ASSERT(tsk == cur_task);
  ASSERT(fd >= 0);

  if (tsk->state != TSK_STOPPED)
    return -1;

  if (tsk->redirs.u >= tsk->redirs.a)
    {
      tsk->redirs.a += 16;
      tsk->redirs.v = xrealloc(tsk->redirs.v,
                               tsk->redirs.a * sizeof(tsk->redirs.v[0]));
    }
  i = tsk->redirs.u++;

  tsk->redirs.v[i].fd = fd;
  tsk->redirs.v[i].tag = mode;

  va_start(args, mode);
  switch (mode)
    {
    case TSR_DUP:
      tsk->redirs.v[i].u.fd2 = va_arg(args, int);
      break;
    case TSR_FILE:
      tsk->redirs.v[i].u.s.path = xstrdup(va_arg(args, char*));
      tsk->redirs.v[i].u.s.oflag = va_arg(args, int);
      tsk->redirs.v[i].u.s.mode = va_arg(args, int);
      /* FIXME: is this necessary */
      tsk->redirs.v[i].u.s.mode |= O_BINARY;
      break;
    case TSR_CLOSE:
      break;
    case TSR_PIPE:
      (void) va_arg(args, int);
      tsk->redirs.u--;
      va_end(args);
      return -1;
    default:
      SWERR(("task_SetRedir: mode == %d", mode));
    }
  va_end(args);

  return 0;
}

  int
task_SetEntryFunction(tTask *tsk, int (*func)(int, char **))
{
  ASSERT(tsk);
  ASSERT(tsk == cur_task);
  tsk->main = func;
  return 0;
}

/**
 * NAME:    print_as_shell_redir
 * PURPOSE: pretty print task file descriptor redirection
 * ARGS:    oflags - open flags specified for redirection
 * RETURN:  string - redirection in external form
 */
  static char *
print_as_shell_redir(int oflags)
{
  if ((oflags & O_RDWR))
    return "<>";
  if ((oflags & O_APPEND))
    return ">>";
  if ((oflags & O_WRONLY))
    return ">";
  if ((oflags & O_RDONLY))
    return "<";
  return "?";
}

/**
 * NAME:    task_PrintArgs
 * PURPOSE: print arguments of the task, including redirections
 * ARGS:    tsk - task
 * RETURN:  0 - ok, -1 - error
 */
  int
task_PrintArgs(tTask *tsk)
{
  int i;

  ASSERT(tsk);
  ASSERT(tsk == cur_task);

  if (tsk->state != TSK_STOPPED)
    return -1;

  if (!tsk->path && tsk->args.u > 0)
    tsk->path = xstrdup(tsk->args.v[0]);

  if (verbose_flag)
    {
      if (tsk->main) {
        fprintf(stderr, "task_Start: 0x%08x(%d):",
                (unsigned int) tsk->main, tsk->args.u);
      } else {
        fprintf(stderr, "task_Start: execv(%d):", tsk->args.u);
      }
      for (i = 0; i < tsk->args.u; i++)
        fprintf(stderr, " %s", tsk->args.v[i]?tsk->args.v[i]:"<NULL>");

      for (i = 0; i < tsk->redirs.u; i++)
        {
          fprintf(stderr, " %d", tsk->redirs.v[i].fd);
          switch (tsk->redirs.v[i].tag)
            {
            case TSR_FILE:
              fprintf(stderr, "%s%s",
                      print_as_shell_redir(tsk->redirs.v[i].u.s.oflag),
                      tsk->redirs.v[i].u.s.path);
              break;
            case TSR_CLOSE:
              fprintf(stderr, "-");
              break;
            case TSR_DUP:
              fprintf(stderr, ">&%d", tsk->redirs.v[i].u.fd2);
              break;
            default:
              SWERR(("task_Start: invalid redirection %d",
                     tsk->redirs.v[i].tag));
            }
        }

      fprintf(stderr, "\n");
    }
  return 0;
}

/**
 * NAME:    task_Start
 * PURPOSE: start a task
 * ARGS:    tsk - task
 * RETURN:  0 - ok, -1 - error
 */
  int
task_Start(tTask *tsk)
{
  int     i;
  tRedir *rdr;
  int     tfd;
  int     sfd;
  int     max_fd = 128;
  int     fd_map[128];
  int     close_flag[128];
  int     code = -1;
  int     saved_statflags;
  int     result;

  ASSERT(tsk);
  ASSERT(tsk == cur_task);

  if (tsk->state != TSK_STOPPED)
    return -1;

  if (!tsk->path && tsk->args.u > 0)
    tsk->path = xstrdup(tsk->args.v[0]);

  if (verbose_flag) task_PrintArgs(tsk);

  if (tsk->main) {
    int code;

    /* just call the specified function */
    tsk->state = TSK_RUNNING;
    code = tsk->main(tsk->args.u, tsk->args.v);
    tsk->state = TSK_EXITED;
    tsk->is_exited = 1;
    tsk->exit_code = code;
    return 0;
  }

  /* FIXME: we assume here, that we have more, than 15
   * available file descriptors and we copy the low ones
   * to that with numbers 15 and more */
  for (i = 0; i < max_fd; i++) {
    close_flag[i] = 0;
    fd_map[i] = -1;
  }
  sfd = 25;

  /* perform redirections */
  for (i = 0, rdr = tsk->redirs.v; i < tsk->redirs.u; i++, rdr++)
    {
      struct stat tmp;

      /* save the current file descriptor */
      saved_statflags = _djstat_flags;
      _djstat_flags = 0;
      result = fstat(tsk->redirs.v[i].fd, &tmp);
      _djstat_flags = saved_statflags;
      if (result >= 0) {
        errno = 0;
        if (dup2(tsk->redirs.v[i].fd, sfd) < 0) {
          write_log(LOG_REUSE, LOG_CRIT,
                    "task_Start: failed to save fd %d, %d: %s",
                    tsk->redirs.v[i].fd, sfd,
                    os_ErrorString());
          goto _cleanup;
        }
        fd_map[tsk->redirs.v[i].fd] = sfd;
        //fprintf(stderr, "saved fd %d as %d\n",
        //        tsk->redirs.v[i].fd, sfd);
        sfd++;
        //close(tsk->redirs.v[i].fd);
      }

      switch (tsk->redirs.v[i].tag)
        {
          case TSR_FILE:
            errno = 0;
            if ((tfd = open(tsk->redirs.v[i].u.s.path,
                            tsk->redirs.v[i].u.s.oflag,
                            tsk->redirs.v[i].u.s.mode)) < 0)
              {
                write_log(LOG_REUSE, LOG_CRIT,
                          "task_Start: failed to open(%s, %d, %4.4o): %s",
                          tsk->redirs.v[i].u.s.path,
                          tsk->redirs.v[i].u.s.oflag,
                          tsk->redirs.v[i].u.s.mode,
                          os_GetErrorString(errno));
                goto _cleanup;
              }

            errno = 0;
            if (dup2(tfd, tsk->redirs.v[i].fd) != tsk->redirs.v[i].fd)
              {
                write_log(LOG_REUSE, LOG_CRIT,
                          "task_Start: failed to dup2(%d, %d): %s",
                          tfd, tsk->redirs.v[i].fd,
                          os_GetErrorString(errno));
                goto _cleanup;
              }
            close(tfd);
            write(tsk->redirs.v[i].fd, "/*xxx*/\n", 8);
            close_flag[tsk->redirs.v[i].fd] = 1;
            break;
          case TSR_DUP:
            errno = 0;
            if (dup2(tsk->redirs.v[i].u.fd2, tsk->redirs.v[i].fd) < 0)
              {
                write_log(LOG_REUSE, LOG_CRIT,
                          "task_Start: failed to dup2(%d, %d): %s",
                          tsk->redirs.v[i].u.fd2,
                          tsk->redirs.v[i].fd,
                          os_GetErrorString(errno));
                goto _cleanup;
              }
            close_flag[tsk->redirs.v[i].fd] = 1;
            break;
          case TSR_CLOSE:
            errno = 0;
            if (close(tsk->redirs.v[i].fd) < 0)
              {
                write_log(LOG_REUSE, LOG_ERR,
                          "task_Start: failed to close(%d): %s",
                          tsk->redirs.v[i].fd,
                          os_GetErrorString(errno));
              }
            break;
          default:
            write_log(LOG_REUSE, LOG_CRIT, 
                      "task_Start: child: invalid redirection %d",
                      tsk->redirs.v[i].tag);
            goto _cleanup;
          }
      }

  errno = 0;
  result = spawnv(P_WAIT, tsk->path, tsk->args.v);
  if (result == -1) {
    write_log(LOG_REUSE, LOG_CRIT,
              "task_Start: cannot spawn a process: %s",
              os_ErrorString());
    goto _cleanup;
  }
  tsk->code = result;
  tsk->state = TSK_EXITED;

  code = 0;

 _cleanup:
  for (i = 0; i < max_fd; i++) {
    if (close_flag[i]) close(i);
    if (fd_map[i] < 0) continue;
    if (fd_map[i] == i) continue;

    if (dup2(fd_map[i], i) < 0) {
      SWERR(("task_Start: cannot restore fd: %d, %d", i, fd_map[i]));
    }
    close(fd_map[i]);
    //fprintf(stderr, "restored fd %d from %d\n",
    //        i, fd_map[i]);
  } 
  return code;
}

/**
 * NAME:    task_Wait
 * PURPOSE: wait for a specified task
 * ARGS:    tsk - task
 * RETURN:  the task pointer itself, or NULL if error
 */
  tTask *
task_Wait(tTask *tsk)
{
  ASSERT(tsk);
  ASSERT(tsk == cur_task);
  ASSERT(tsk->state != TSK_RUNNING);

  if (tsk->state == TSK_ERROR || tsk->state == TSK_STOPPED)
    return 0;
  return tsk;
}

/**
 * NAME:    task_Status
 * PURPOSE: return task status
 * ARGS:    tsk - task
 * RETURN:  task status
 */
  int
task_Status(tTask *tsk)
{
  ASSERT(tsk);
  ASSERT(tsk == cur_task);
  return tsk->state;
}

/**
 * NAME:    task_TermSignal
 * PURPOSE: return termination signal
 * ARGS:    tsk - task
 * RETURN:  termination signal, or -1 if task is not terminated by signal
 */
  int
task_TermSignal(tTask *tsk)
{
  ASSERT(tsk);
  ASSERT(tsk == cur_task);
  return (unsigned) tsk->code >> 8; 
}

/**
 * NAME:    task_ExitCode
 * PURPOSE: get exit code of the task
 * ARGS:    tsk - task
 * RETURN:  exit code of the task, or -1 is task is not exited
 */
  int
task_ExitCode(tTask *tsk)
{
  ASSERT(tsk);
  ASSERT(tsk == cur_task);
  if (tsk->state != TSK_EXITED)
    return -1;
  if (tsk->is_exited)
    return tsk->exit_code;
  return (tsk->code & 0xFF);
}

/**
 * NAME:    task_IsAbnormal
 * PURPOSE: check for abnormal task termination
 * ARGS:    tsk - task
 * RETURN:  0 - task exited with code 0,
 *          1 - task either terminated with signal, or exited with code > 0
 */
  int
task_IsAbnormal(tTask *tsk)
{
  ASSERT(tsk);
  ASSERT(tsk == cur_task);
  if (tsk->state == TSK_SIGNALED)
    return 1;
  if (tsk->state == TSK_EXITED && tsk->is_exited && tsk->exit_code > 0)
    return 1;
  if (tsk->state == TSK_EXITED && tsk->code > 0)
    return 1;
  return 0;
}

/**
 * NAME:    task_Log
 * PURPOSE: log task termination event
 * ARGS:    tsk - task
 *          fac - logging facility
 *          sev - logging severity
 */
  void
task_Log(tTask *tsk, int fac, int sev)
{
  ASSERT(tsk);
  ASSERT(tsk == cur_task);

  if (tsk->state != TSK_SIGNALED && tsk->state != TSK_EXITED)
    return;
  if (tsk->state == TSK_SIGNALED)
    {
      write_log(fac, sev,
                "process is terminated with signal %d (%s)",
                REUSE_TERMSIG(tsk->code),
                os_GetSignalString(REUSE_TERMSIG(tsk->code)));
    }
  else
    {
      int   r = REUSE_EXITSTATUS(tsk->code);
      if (tsk->is_exited)
        r = tsk->exit_code;
      write_log(fac, sev, "process is exited with code %d", r);
    }
}

/* === Do not add stuff after this line === */

/*
 * Local variables:
 *  compile-command: "make -C .."
 *  c-font-lock-extra-types: ("FILE" "\\sw+_t" "tRedir" "tTask" "va_list")
 * End:
 */
