/*$Id: logger.c 57 2001-06-23 07:04:26Z cher $*/

/* Copyright (C) 2000-2001 Alexander Chernov <cher@ispras.ru> */
/* Created: <2000-07-01 19:53:32 cher> */

/*
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 */

/**
 * FILE:    utils/logger.c
 * PURPOSE: logging facilities, fatal error handling
 */

static char const rcsid[]=
"$Id: logger.c 57 2001-06-23 07:04:26Z cher $";

#include "lconfig.h"
#include <p_config.h>

#include CONF_DEPFILE("reuse_version.h")

#include "reuse/logger.h"
#include "reuse/xalloc.h"
#include "reuse/osdeps.h"

#include <stdio.h>
#include <stdarg.h>
#include <assert.h>
#include <unistd.h>
#include <time.h>
#include <fcntl.h>
#include <sys/stat.h>

static int   log_fd    = 2;     /* logging file descriptor */
static char *log_path  = NULL;  /* full path to the log file */

static logmodule_t default_module =
{ 0, "", LOG_INFO };
static logmodule_t logger_module =
{ LOG_LOGGER, "logger", LOG_ERR };
static logmodule_t sw_module =
{ LOG_SW, 0, LOG_ERR };
static logmodule_t utils_module =
{ LOG_REUSE, "reuse", LOG_INFO };
static logmodule_t *logmodules[LOG_MAX_MODULE_NUM];

#define MAX_LOG_LEVEL 50

#define LOG_MIN_PRIO LOG_JUNK
#define LOG_MAX_PRIO LOG_EMERG
static char *priority_names[]=
{
  "junk", "debug", "info", "notice", "warning",
  "error", "critical", "alert", "emerg"
};

static int initialized = 0;
static void minimal_init(void)
{
  if (initialized) return;
  initialized = 1;

  if (log_fd < 0) log_fd = 2;
  log_path = NULL;

  memset(logmodules, 0, sizeof(logmodules));

  logmodules[0] = &default_module;
  logmodules[LOG_LOGGER] = &logger_module;
  logmodules[LOG_SW] = &sw_module;
  logmodules[LOG_REUSE] = &utils_module;
}

/**
 * NAME:    logger_init
 * PURPOSE: initialize the logger module
 */
  void
logger_init(logmodule_t *mi, char *path)
{
  int i;

  minimal_init();

  if (mi) {
    for (i = 0; mi[i].name; i++) {
      ASSERT(mi[i].num > 3 && mi[i].num < LOG_MAX_MODULE_NUM);
      logmodules[mi[i].num] = &mi[i];
    } /* for (i) */
  } /* if (mi) */

  if (path) {
    if ((i = open(path, O_BINARY|O_WRONLY|O_CREAT|O_APPEND, S_IWUSR)) < 0) {
      fprintf(stderr, "Cannot open log file '%s'\n", path);
    } else {
      log_path = path;
      log_fd   = i;
    }
  }
}

/**
 * NAME:    vwrite_log
 * PURPOSE: write log message
 * ARGS:    facility - logging facility
 *          level    - severity level
 *          format   - user provided message
 *          args     - extra format specific arguments
 * RETURN:  number of bytes written to the log file
 */
  int
vwrite_log(int facility, int level, char *format, va_list args)
{
  int        r;
  time_t     tt;
  struct tm *ptm;
  char      *atm;
  char      *prio, bprio[32];
  char      *pfac, bfac[32];
  int        msglen = 1023;
  char       msg[1024];

  assert (format != NULL);
  minimal_init();

  /* check whether log file is open */
  if (log_fd <= 0) return 0;

  /* check for valid module */
  if (facility < 0 || facility >= LOG_MAX_MODULE_NUM) facility = 0;
  if (!logmodules[facility]) facility = 0;

  /* check that the facility is blocked */
  if (logmodules[facility]->blocked) return 0;

  /* check against minimal log level for the facility */
  if (level < logmodules[facility]->level) return 0;

  time(&tt);
  ptm = localtime(&tt);
  atm = asctime(ptm);
  atm[strlen(atm) - 1] = 0;

  if (level < LOG_MIN_PRIO || level > LOG_MAX_PRIO) {
    sprintf(bprio, "%d", level);
    prio = bprio;
  } else {
    prio = priority_names[level];
  }

  if (!logmodules[facility]->name) {
    sprintf(bfac, "%d", facility);
    pfac = bfac;
  } else {
    pfac = logmodules[facility]->name;
  }

  if (facility == 0) {
    r = sprintf(msg, "%s:%s:", atm, prio);
  } else {
    r = sprintf(msg, "%s:%s:%s:", atm, pfac, prio);
  }

  
  os_vsnprintf(msg + r, msglen - r, format, args);
  msg[msglen] = 0;
  r = strlen(msg);
  msg[r++] = '\n';
  
  if (r >= 1024) {
    fprintf(stderr, "fatal buffer overrun in logger module\n");
    abort();
  }

  if (write(log_fd, msg, r) < 0) {
    fprintf(stderr, "log file write error: %s\n", os_ErrorString());
    fprintf(stderr, "closing log file\n");
    close(log_fd);
    log_fd = -1;
    return -1;
  }

  return r;
}

/**
 * NAME:    write_log
 * PURPOSE: write log message
 * ARGS:    facility - logging facility
 *          level    - severity level
 *          format   - user provided message
 *          ...      - extra format specific arguments
 * RETURN:  number of bytes written to the log file
 */
  int
write_log(int facility, int level, char *format, ...)
{
  va_list    args;
  int        r;

  va_start(args, format);
  r = vwrite_log(facility, level, format, args);
  va_end(args);
  return r;
}

/**
 * NAME:    logger_close
 * PURPOSE: close log file
 */
  void
logger_close(void)
{
  minimal_init();
}

/**
 * NAME:    logger_get_fd
 * PURPOSE: get logging file descriptor
 * RETURN:  logging file descriptor
 */
  int
logger_get_fd(void)
{
  minimal_init();
  return log_fd;
}

/* ==================== Fatal errors handling ======================= */

/* hope the message won't be that long */
#define FATALBUFLEN 1024
static char fatalbuf[FATALBUFLEN]; /* buffer for the fatal error message */

static char    *swerr_file;     /* source file name */
static int      swerr_line;     /* source file line */

#ifdef __GNUC__
static void _swerr(char *file, int line, char *format, va_list args) __attribute__ ((noreturn));
#else
static void _swerr();
#endif

/**
 * NAME:    _swwarn
 * PURPOSE: report non-fatal software error (warning)
 * ARGS:    file   - source file name
 *          line   - source file line
 *          format - message text
 *          args   - extra format specific arguments
 */
  static void
_swwarn(char *file, int line, char *format, va_list args)
{
  int   n;
  char *s = fatalbuf;

  n = sprintf(s, "Internal: %s: %d: ", file, line);
  n = vsprintf(s += n, format, args);
  sprintf(s += n, "\n");

  write_log(LOG_SW, LOG_ALERT, "%s", fatalbuf);
}

/**
 * NAME:    _swerr
 * PURPOSE: report fatal software error
 * ARGS:    file   - source file name
 *          line   - source line number
 *          jb     - fatal error recovery point
 *          format - message text
 *          args   - extra format specific arguments
 * RETURN:  
 */
   static void
_swerr(char *file, int line, char *format, va_list args)
{
  int   n;
  char *s = fatalbuf;

  n = sprintf(s, "Internal: %s: %d: ", file, line);
  n = vsprintf(s += n, format, args);
  sprintf(s += n, "\n");

  write_log(LOG_SW, LOG_EMERG, "%s", fatalbuf);

  /* one day we'll do longjmp */
  //abort();
  // let's do exit, on dos we don't have coredumps anyway
  exit(3);
}

/**
 * NAME:    swerr
 * PURPOSE: report software fatal error
 * ARGS:    file   - source file name
 *          line   - source line number
 *          format - error message
 *          ...    - extra format specific arguments
 */
  void
swerr(char *file, int line, char *format, ...)
{
  va_list args;

  va_start(args, format);
  _swerr(file, line, format, args);
}

/**
 * NAME:    swerr1
 * PURPOSE: report software fatal error
 * ARGS:    format - error message
 *          ...    - extra message specific arguments
 * NOTE:    error recovery is performing using the default error recovery point
 */
  void
swerr1(char *format, ...)
{
  va_list args;

  va_start(args, format);
  _swerr(swerr_file, swerr_line, format, args);
}

/**
 * NAME:    swerr2
 * PURPOSE: report software fatal error
 * ARGS:    format - error message
 *          ...    - extra message specific arguments
 * NOTE:    error recovery is performing using the user error recovery point
 */
  void
swerr2(char *format, ...)
{
  va_list args;

  va_start(args, format);
  _swerr(swerr_file, swerr_line, format, args);
}

/**
 * NAME:    swerr_SetPos
 * PURPOSE: set source position to report fatal error later with swerr1
 * ARGS:    file - source file name
 *          line - source line number
 */
  void
swerr_SetPos(char *file, int line)
{
  swerr_file = file;
  swerr_line = line;
}

/**
 * NAME:    swerr_SetPos
 * PURPOSE: set source position and recovery point
 *          to report fatal error later with swerr2
 * ARGS:    file - source file name
 *          line - source line number
 *          buf  - pointer to jmp_buf - error recovery point
 */
  void
swerr_SetPosBuf(char *file, int line, void *buf)
{
  swerr_file    = file;
  swerr_line    = line;

  /* This does not work : swerr_handler = *((jmp_buf) buf); */
  //memcpy(&swerr_handler, buf, sizeof (swerr_handler));
}

/**
 * NAME:    swabort
 * PURPOSE: software abort
 * NOTE:    this should normally perform fatal error recovery
 */
  void
swabort(void)
{
  exit(3);
  //abort();
}

/**
 * NAME:    swwarn
 * PURPOSE: write software warning (non fatal error)
 * ARGS:    format - message
 *          ...    - extra message specific arguments
 * NOTE:    source position should be set separately by swerr_SetPos
 */
  void
swwarn(char *format, ...)
{
  va_list args;

  va_start(args, format);
  _swwarn(swerr_file, swerr_line, format, args);
}

/* === Do not add stuff after this line === */

/*
 * Local variables:
 *  compile-command: "make -C .."
 *  c-font-lock-extra-types: ("FILE" "\\sw+_t" "va_list" "jmp_buf")
 * End:
 */
