/* $Id: osdeps.c 57 2001-06-23 07:04:26Z cher $ */

/* Copyright (C) 2000-2001 Alexander Chernov <cher@ispras.ru> */
/* Created: <2000-07-01 19:58:56 cher> */

/*
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 */

/**
 * FILE:    utils/osdeps.c
 * PURPOSE: miscellanious os-dependent functions
 */

/**
 * The following functions are not implemented for ix86-msdos-djgpp:
 *   os_MakeDirPath, os_MakeDir, os_SetLock.
 */

static char const cvsid[] =
"$Id: osdeps.c 57 2001-06-23 07:04:26Z cher $";

#include "lconfig.h"
#include <p_config.h>

#include CONF_DEPFILE("reuse_version.h")

#include "reuse/osdeps.h"
#include "reuse/xalloc.h"
#include "reuse/logger.h"

#define NOT_IMPLEMENTED() SWERR(("%s: not implemented", __FUNCTION__))

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <assert.h>

#include <unistd.h>
#include <sys/utsname.h>
#include <sys/stat.h>
#include <signal.h>

#include <dos.h>
#include <dir.h>

static char errbuf[256];
/**
 * NAME:    os_GetErrorString
 * PURPOSE: get error string by error code
 * ARGS:    e - error code
 * RETURN:  the corresponding error string
 */
  char *
os_GetErrorString(int e)
{
  char *s = strerror(e);
  if (s == NULL)
    {
      sprintf(errbuf, "unknown error (%d)", e);
      return errbuf;
    }
  return s;
}

/**
 * NAME:    os_ErrorString
 * PURPOSE: get error string of the current error (errno variable)
 * RETURN:  error string for the current error
 */
  char *
os_ErrorString(void)
{
  return os_GetErrorString(errno);
}

/**
 * NAME:    os_GetSignalString
 * PURPOSE: get the signal name
 * ARGS:    s - signal number
 * RETURN:  string with the signal name
 */
  char *
os_GetSignalString(int s)
{
  static char tmpbuf[32];

  if (s <= 0 || s >= NSIG) {
    sprintf(tmpbuf, "unknown signal %d", s);
    return tmpbuf;
  }
  return sys_siglist[s];
}

/**
 * NAME:    dos_path_special_prefix
 * PURPOSE: check special dos path prefixes (a:\, a:, \)
 * ARGS:    path - dos path
 * RETURN:  NULL - no special prefix
 *          else special prefix in canonical form
 * NOTE:    string 'path' is modified
 */
  static char*
dos_path_special_prefix(char *path)
{
  int i;

  if (!path) return NULL;

  if (isalpha((unsigned char) path[0])
      && path[1] == ':'
      && (path[2] == '/' || path[2] == '\\')) {
    for (i = 3; path[i] == '/' || path[i] == '\\'; i++);
    if (path[i] == 0) {
      path[0] = tolower(path[0]);
      path[2] = '/';
      path[3] = 0;
      return path;
    }
  }
  if (isalpha((unsigned char) path[0])
      && path[1] == ':'
      && path[2] != '/' && path[2] != '\\' && !isalpha(path[2])) {
    path[0] = tolower((unsigned char) path[0]);
    path[2] = 0;
    return path;
  }
  if (path[0] == '/' || path[0] == '\\') {
    for (i = 1; path[i] == '/' || path[i] == '\\'; i++);
    if (path[i] == 0) {
      path[0] = '/';
      path[1] = 0;
      return path;
    }
  }

  return NULL;
}

char *os_GetWorkingDir(void);

  static char*
dos_canonize_path(char const *name)
{
  char *path = 0;

  if (!name) return NULL;
  XCALLOC(path, FILENAME_MAX);
  _fixpath(name, path);
  return path;
}

/**
 * NAME:    os_FindInPath
 * PURPOSE: find the given executable in the PATH
 * ARGS:    name - name of the executable
 * RETURN:  full path to the file, or NULL if file not found
 */
  char *
os_FindInPath(char *name)
{
  char *res = searchpath(name);
  if (!res) return NULL;
  return xstrdup(res);
}

#define EXTBUFSIZE 32
static char extbuf[EXTBUFSIZE];

/**
 * NAME:    os_GetSuffix
 * PURPOSE: get file suffix
 * ARGS:    path - path of the file
 * RETURN:  file suffix in the static location
 */
  char *
os_GetSuffix(char *path)
{
  os_rGetSuffix(path, extbuf, EXTBUFSIZE);
  return extbuf;
}

/**
 * NAME:    os_rGetSuffix
 * PURPOSE: get the suffix of the file (reentrant version)
 * ARGS:    path - path of the file
 *          buf  - buffer to store suffix
 *          size - size of the buffer
 * RETURN:  length of the suffix string
 */
  int
os_rGetSuffix(char *path, char *buf, int size)
{
  int len;
  char *s;

  if (size <= 0) return 0;
  if (!path || !buf) return 0;
  buf[0] = 0;
  if (size == 0) return 0;
  len = strlen(path);
  if (!len) return 0;

  for (s = path + len - 1;
       s >= path && *s != '.' && *s != '/' && *s != '\\' && *s != ':';
       s--);
  if (s < path || *s != '.') return 0;
  if (s == path || s[-1] == '/' || s[-1] == '\\' || s[-1] == ':') return 0;

  strncpy(buf, s, size);
  buf[size - 1] = 0;
  return strlen(buf);
}

/**
 * NAME:    os_GetBasename
 * PURPOSE: get basename of the file
 * ARGS:    str - path of the file
 * RETURN:  basename of the file, allocated in the heap
 */
  char *
os_GetBasename(char *str)
{
  char *path = xstrdup(str);
  char *p;
  int   i;
  
  if ((p = dos_path_special_prefix(path)) != NULL) {
    xfree(path);
    return xstrdup("");
  }

  // remove trailing slashes
  for (i = strlen(path) - 1;
       i >= 0 && (path[i] == '/' || path[i] == '\\');
       i--) {
    path[i] = 0;
  }

  for (; i >= 0 && path[i] != '/' && path[i] != '\\' && path[i] != ':'; i--);

  p = xstrdup(path + i + 1);
  xfree(path);
  return p;
}

/**
 * NAME:    os_SubstSuffix
 * PURPOSE: replace the suffix of the path with the given new suffix
 * ARGS:    path   - path of the file
 *          suffix - new suffix
 * RETURN:  new path allocated in the heap
 */
  char *
os_SubstSuffix(char *path, char *suffix)
{
  int   slen;
  char  buf[16];
  int   i;
  char *p;

  if (!path) return xstrdup(suffix);
  if (!suffix) return xstrdup(path);

  slen = os_rGetSuffix(path, buf, 16);
  if (slen == 0) {
    return xstrmerge2(path, suffix);
  }

  for (i = strlen(path) - 1; path[i] != '.'; i--);
  assert(i > 0);

  XCALLOC(p, (i) + strlen(suffix) + 2);
  memcpy(p, path, i);
  path[i] = 0;
  strcat(p, suffix);
  return p;
}

#if 0
/**
 * NAME:    strip_trailing_slashes
 * PURPOSE: Remove trailing slashes from PATH.
 * ARGS:    path -- directory path 
 *         
 * RETURN:  void
 * NOTE: 
   This is useful when using filename completion from a shell that
   adds a "/" after directory names (such as tcsh and bash), because
   the Unix rename and rmdir system calls return an "Invalid argument" error
   when given a path that ends in "/" (except for the root directory).  
 */
static void
strip_trailing_slashes (path)
     char *path;
{
  /*
  int last;

  last = strlen (path) - 1;
  while (last > 0 && path[last] == '/')
    path[last--] = '\0';
  */
  NOT_IMPLEMENTED();
}
#endif

#if 0
/**
 * NAME:    make_path
 * PURPOSE: make directory hierarchy
 * ARGS:     
 *         
 * RETURN:  1 -- error, 0 -- ok 
 * NOTE: 
 * Ensure that the directory ARGPATH exists.

   Create any leading directories that don't already exist, with
   permissions PARENT_MODE.
   If the last element of ARGPATH does not exist, create it as
   a new directory with permissions MODE.
   If OWNER and GROUP are non-negative, use them to set the UID and GID of
   any created directories.
   If PRESERVE_EXISTING is non-zero and ARGPATH is an existing directory,
   then do not attempt to set its permissions and ownership.

   Return 0 if ARGPATH exists as a directory with the proper
   ownership and permissions when done, otherwise 1.  */

static int
make_path (const char *argpath,
           int mode,
           int parent_mode,
           uid_t owner,
           gid_t group,
           int preserve_existing)
{
#if 0
  char *dirpath;                /* A copy we can scribble NULs on.  */
  struct stat stats;
  int retval = 0;
  int oldmask = umask (0);
  dirpath = (char *) xmalloc (strlen (argpath) + 1);
  strcpy (dirpath, argpath);
  strip_trailing_slashes(dirpath); /* del last '/' */
  if (stat (dirpath, &stats))
    {
      char *slash;
      int tmp_mode;             /* Initial perms for leading dirs.  */
      int re_protect;           /* Should leading dirs be unwritable? */
      struct ptr_list
      {
        char *dirname_end;
        struct ptr_list *next;
      };
      struct ptr_list *p, *leading_dirs = NULL;

      /* If leading directories shouldn't be writable or executable,
         or should have set[ug]id or sticky bits set and we are setting
         their owners, we need to fix their permissions after making them.  */


      if (((parent_mode & 0300) != 0300)
          || (owner != (uid_t) -1 && group != (gid_t) -1
              && (parent_mode & 07000) != 0))
        {
          tmp_mode = 0700;
          re_protect = 1;
        }
      else
        {
          tmp_mode = parent_mode;
          re_protect = 0;
        }

      slash = dirpath;
      while (*slash == '/')
        slash++;
      while ((slash = strchr (slash, '/')))
        {
          *slash = '\0';
          if (stat (dirpath, &stats))
            {
              if (mkdir (dirpath, tmp_mode))
                {
                  //error (0, errno, "cannot create directory `%s'", dirpath);
                  retval=1;
      goto ret;
                }
              else
                {

                  if (owner != (uid_t) -1 && group != (gid_t) -1
                      && chown (dirpath, owner, group)
                      )
                    {
                      //error (0, errno, "%s", dirpath);
                      retval = 1;
                    }
                  if (re_protect)
                    {
                      struct ptr_list *new = (struct ptr_list *)
                        alloca (sizeof (struct ptr_list));
                      new->dirname_end = slash;
                      new->next = leading_dirs;
                      leading_dirs = new;
                    }
                }
            }
          else if (!S_ISDIR (stats.st_mode))
            {
              //error (0, 0, "`%s' exists but is not a directory", dirpath);
        errno=20;
        retval=1;
        goto ret;
            }

          *slash++ = '/';

          /* Avoid unnecessary calls to `stat' when given
             pathnames containing multiple adjacent slashes.  */
          while (*slash == '/')
            slash++;
        }

      /* We're done making leading directories.
         Create the final component of the path.  */

      /* The path could end in "/." or contain "/..", so test
         if we really have to create the directory.  */

      if (stat (dirpath, &stats) && mkdir (dirpath, mode))
        {
          //error (0, errno, "cannot create directory `%s'", dirpath);
    retval=1;
    goto ret;
        }

      if (owner != (uid_t) -1 && group != (gid_t) -1)
        {
          if (chown (dirpath, owner, group))
            {
              //error (0, errno, "%s", dirpath);
              retval = 1;
            }
          /* chown may have turned off some permission bits we wanted.  */
          if ((mode & 07000) != 0 && chmod (dirpath, mode))
            {
              //error (0, errno, "%s", dirpath);
              retval = 1;
            }
        }

      /* If the mode for leading directories didn't include owner "wx"
         privileges, we have to reset their protections to the correct
         value.  */
      for (p = leading_dirs; p != NULL; p = p->next)
        {
          *(p->dirname_end) = '\0';
          if (chmod (dirpath, parent_mode))
            {
              //error (0, errno, "%s", dirpath);
              retval = 1;
            }
        }
    }
  else
    {
      /* We get here if the entire path already exists.  */

      if (!S_ISDIR (stats.st_mode))
        {
          //error (0, 0, "`%s' exists but is not a directory", dirpath);
    errno=20;
    retval=1;
    goto ret;
        }

      if (!preserve_existing)
        {
          /* chown must precede chmod because on some systems,
             chown clears the set[ug]id bits for non-superusers,
             resulting in incorrect permissions.
             On System V, users can give away files with chown and then not
             be able to chmod them.  So don't give files away.  */

          if (owner != (uid_t) -1 && group != (gid_t) -1
              && chown (dirpath, owner, group))
            {
              //error (0, errno, "%s", dirpath);
              retval = 1;
            }
          if (chmod (dirpath, mode))
            {
              //error (0, errno, "%s", dirpath);
              retval = 1;
            }
        }
    }
ret:
//  fprintf(stderr, "exiting with errno=%d\n", errno);
  xfree(dirpath);
  umask (oldmask);
  return retval;
#endif
  NOT_IMPLEMENTED();
}
#endif



/**
 * NAME:    os_MakeDirPath
 * PURPOSE: make directory hierarchy
 * ARGS:    path - path to the directory to be created
 *          mode - directory mode
 * RETURN:  <0 - error, >= 0 - ok
 * NOTE:    not implemented
 */
  int
os_MakeDirPath(char *path, int mode)
{
#if 0
  int newmode, parent_mode, err;
  int old_mask = umask(0);
  newmode = 0777 & ~old_mask;
  parent_mode = newmode | 0300; /* u+wx */
   
  errno=0;
  err = make_path (path, newmode, parent_mode,
                     -1, -1, 1);
  umask(old_mask);
  return -err;
#endif
  NOT_IMPLEMENTED();
}

/**
 * NAME:    os_MakeDir
 * PURPOSE: create directory
 * ARGS:    path - path to the directory
 *          mode - directory permissions mode
 * RETURN:  <0 - error, >0 - ok, 1 - directory already exists
 */
  int
os_MakeDir(char *path, int mode)
{
  /*
  struct stat buf;
  int         r;

  if (!path) return -1;

  errno = 0;
  r = stat(path, &buf);
  if (r < 0 && errno != ENOENT) return r;
  if (r >= 0 && !S_ISDIR(buf.st_mode)) return -1;
  if (r >= 0) return 1;

  return mkdir(path, mode);
  */
  NOT_IMPLEMENTED();
}

/**
 * NAME:    os_DirName
 * PURPOSE: return dir name from file path
 * ARGS:    argpath - path to the file
 * RETURN:  the directory component of the path
 */
  char *
os_DirName(char const *str)
{
  char *path = xstrdup(str);
  char *p;
  int   len = strlen(path);
  int   i;

  if (!strcmp(path, "")) {
    xfree(path);
    return xstrdup(".");
  }

  // consider special cases
  if ((p = dos_path_special_prefix(path)) != NULL) {
    return p;
  }

  // remove trailing slashes
  for (i = len - 1; i >= 0 && (path[i] == '/' || path[i] == '\\'); i--) {
    path[i] = 0;
  }

  // ignore name component
  for (; i >= 0 && path[i] != '/' && path[i] != '\\' && path[i] != ':'; i--) {
    path[i] = 0;
  }

  // consider special cases again
  if ((p = dos_path_special_prefix(path)) != NULL) {
    return p;
  }

  // remove trailing slashes
  for (; i >= 0 && (path[i] == '/' || path[i] == '\\'); i--) {
    path[i] = 0;
  }

  if (!i) {
    xfree(path);
    return xstrdup(".");
  }
  return path;
}

/**
 * NAME:    os_NodeName
 * PURPOSE: get the host name
 * RETURN:  hostname - string in the static location
 */
static struct utsname uname_buf;
  char *
os_NodeName(void)
{
  if (uname(&uname_buf) < 0)
    return 0;

  return uname_buf.nodename;
}

/**
 * NAME:    os_IsFile
 * PURPOSE: check whether the object given by the path is file
 * ARGS:    path - path to the object
 * RETURN:  <0 - error,
 *          OSPK_REG    - regular file
 *          OSPK_DIR    - directory
 *          OSPK_OBJECT - other object
 */
  int
os_IsFile(char *path)
{
  struct stat buf;

  if (!path) return -1;

  if (stat(path, &buf) < 0)
    {
      /* file does not exist */
      return -1;
    }

  if (S_ISREG(buf.st_mode))
    return OSPK_REG;
  if (S_ISDIR(buf.st_mode))
    return OSPK_DIR;
  return OSPK_OTHER;
}

/**
 * NAME:    os_tempnam
 * PURPOSE: wrapper over tempnam function
 * ARGS:    dir - temporary files directory
 *          pfx - file prefix
 * RETURN:  temporary file name allocated in the heap
 */
        char *
os_tempnam(char const *dir, char const *pfx)
{
        return tempnam(dir, pfx);
}

/**
 * NAME:    os_readint
 * PURPOSE: read int value from the string
 * ARGS:    str - string to read from
 *          val - pointer to the location to store read value
 * RETURN:  <0 - error, 0 - ok
 */
        int
os_readint(char const *str, int *val)
{
  char    *s;
  int      t;

  errno = 0;
  t = strtol(str, &s, 10);
  if (*s != 0) return -1;
  if (errno != 0) return -errno;
  *val = t;
  return 0;
}

  int
os_readdouble(char const *str, double *val)
{
  char   *s;
  double  t;

  errno = 0;
  t = strtod(str, &s);
  if (*s != 0) return -1;
  if (errno != 0) return -errno;
  *val = t;
  return 0;
}

/**
 * NAME:    os_GetWorkingDir
 * PURPOSE: Get the current working directory
 * RETURN:  the current working directory, located in the heap memory
 *          so it must be freed by xfree call
 */
  char *
os_GetWorkingDir(void)
{
  char  buf[2048];

  errno = 0;
  if (getcwd(buf, 2048)) {
    return xstrdup(buf);
  }
  if (errno != ERANGE)
    return 0;

  {
    char buf[16384];
    if (getcwd(buf, 16384)) {
      return xstrdup(buf);
    }
  }
  return 0;
}

  char *
get_PRJ_HOME_name(void)
{
#if defined CONF_ENV_PREFIX
  return CONF_ENV_PREFIX "_HOME";
#else
  return "REP_HOME";
#endif /* CONF_ENV_PREFIX */
}

  char *
get_PRJ_HOME(void)
{
  return getenv(get_PRJ_HOME_name());
}

  char *
get_PRJ_CONFIG_name(void)
{
#if defined CONF_ENV_PREFIX
  return CONF_ENV_PREFIX "_CONFIG";
#else
  return "REP_CONFIG";
#endif /* CONF_ENV_PREFIX */
}

  char *
get_PRJ_CONFIG(void)
{
  return getenv(get_PRJ_CONFIG_name());
}

/**
 * NAME:    os_GuessProjectEnv
 * PURPOSE: guess the values of <project>_HOME and <project>_CONFIG variables
 *          if those are not defined
 * ARGS:    path - path to the program (argv[0])
 * RETURN:  0 - ok, <0 - error
 */
  int
os_GuessProjectEnv(char const *path, char const *project)
{
  char *p_config = 0;
  char *p_home   = 0;

  int   rcode;
  char *path1 = 0;
  char *path2 = 0;
  char *path3 = 0;
  char *base1 = 0;
  char *base2 = 0;
  char *env_str = 0;
  char *rep_home = 0;

  p_config = get_PRJ_CONFIG_name();
  p_home   = get_PRJ_HOME_name();

  /* set CONFIG variable */  
  if (!getenv(p_config)) {
    char *str = 0;

    XCALLOC(str, strlen(CONF_CANONIC_STR) + 32);
    sprintf(str, "%s=%s", p_config, CONF_CANONIC_STR);
    if (putenv(str) < 0) {
      /* FIXME: report error */
      xfree(str);
      return -1;
    }
  }

  if (getenv(p_home) != NULL) {
    return 0;
  }

  path1 = dos_canonize_path(path);

  //fprintf(stderr, "path = %s\n", path1);

  if (!(path2 = os_DirName(path1))) goto failure;
  if (!(base1 = os_GetBasename(path2))) goto failure;
  if (!strcasecmp(base1, "bin")) {
    // started from ${P_HOME}/bin/
    if (!(rep_home = os_DirName(path2))) goto failure;
  }
#if defined CONF_CANONIC_STR
  else if (!strcasecmp(base1, CONF_CANONIC_STR)
           || !strcasecmp(base1, CONF_RUNTIME_STR)) {
    // started from ${P_HOME}/libexec/${CONF_CANONIC_STR}
    if (!(path3 = os_DirName(path2))) goto failure;
    if (!(base2 = os_GetBasename(path3))) goto failure;
    if (strcasecmp(base2, "exec") && strcasecmp(base2, "libexec"))
      goto failure;
    if (!(rep_home = os_DirName(path3))) goto failure;
  }
  /* handle mangled (8.3) filename
   * FIXME: should we check more strictly? */
  else if (!strncasecmp(base1, CONF_CANONIC_STR, 6)
           || !strncasecmp(base1, CONF_RUNTIME_STR, 6)) {
    // started from ${P_HOME}/libexec/${CONF_CANONIC_STR}
    if (!(path3 = os_DirName(path2))) goto failure;
    if (!(base2 = os_GetBasename(path3))) goto failure;
    if (strcasecmp(base2, "exec") && strcasecmp(base2, "libexec"))
      goto failure;
    if (!(rep_home = os_DirName(path3))) goto failure;
  }
#endif /* CONF_CANONIC_STR */
  else {
    goto failure;
  }

  /* now rep_home should contain value of <project>_HOME */
  env_str = xmalloc(strlen(rep_home) + 32);
  sprintf(env_str, "%s=%s", p_home, rep_home);
  putenv(env_str);
  env_str = 0;

  rcode = 0;
  goto free_resources;

 failure:
  rcode = -1;

 free_resources:
  xfree(path1);
  xfree(path2);
  xfree(base1);
  xfree(path3);
  xfree(base2);
  xfree(rep_home);
  xfree(env_str);
  return rcode;
}

/**
 *  NAME:    os_SetLock
 *  PURPOSE: set the lockfile with the specified name
 *  ARGS:    path      - lock file path
 *           perms     - file permissions
 *           left_open - 1, if the function should return the open file
 *                       0, if the function should close the file
 *  RETURN:  >= 0 - ok, -1 - file is locked, -2 - other error
 *  NOTE:    function works reliably over NFS (Network Failure System)
 */
  int
os_SetLock(char const *path, int perms, int left_open)
{
#if 0
  char *dirname   = os_DirName(path);
  char *hostname  = os_NodeName();
  char *uniq_name = 0;
  int   fd        = -1;
  int   code      = 0;
  int   file_flag = 0;
  struct stat buf;

  /* 1. Create the unique file on the same filesystem */
  uniq_name = xmalloc(strlen(dirname) + strlen(hostname) + 32);
  sprintf(uniq_name, "%s/LCK_%s_%d", dirname, hostname, (int) getpid());
  if ((fd = open(uniq_name, O_WRONLY | O_CREAT, perms)) < 0) {
    code = -2;
    goto free_resources;
  }
  file_flag = 1;

  /* 2. Use link(2) to create a link to the lockfile */
  /*    do not use the return value of link call */
  link(uniq_name, path);

  /* 3. Use stat(2) on the unique file to check that the link count
        is increased to 2 */
  if (fstat(fd, &buf) < 0) {
    code = -2;
    goto free_resources;
  }
  if (buf.st_nlink != 2) {
    code = -1;
    goto free_resources;
  }

  /* at this point lock file is created */
  if (left_open) {
    code = fd;
    fd = -1;
  }

 free_resources:
  if (file_flag) unlink(uniq_name);
  free(uniq_name);
  free(dirname);
  if (fd >= 0) close(fd);
  return code;
#endif
  NOT_IMPLEMENTED();
}

  int
os_CheckAccess(char const *path, int perms)
{
  int r_perms = 0;
  int res     = 0;

  if ((perms & REUSE_X_OK)) r_perms |= X_OK;
  if ((perms & REUSE_W_OK)) r_perms |= W_OK;
  if ((perms & REUSE_R_OK)) r_perms |= R_OK;
  if ((perms & REUSE_F_OK)) r_perms |= F_OK;

  res = access(path, r_perms);
  return (res >= 0)?res:-errno;
}

extern int reuse_vsnprintf(char *, int, char const *, va_list);

  int
os_snprintf(char *buf, int buflen, char const *format, ...)
{
  int     r;
  va_list args;

  va_start(args, format);
  r = reuse_vsnprintf(buf, buflen, format, args);
  va_end(args);
  return r;
}

  int
os_vsnprintf(char *buf, int buflen, char *format, va_list args)
{
  return reuse_vsnprintf(buf, buflen, format, args);
}

  void *
os_AttachFILE(int handle, char const *mode)
{
  return (void*) fdopen(handle, mode);
}

/* === Do not add stuff after this line === */

/*
 * Local variables:
 *  compile-command: "make -C .."
 * End:
 */
