/* $Id: Printf.h 352 2004-07-17 06:11:56Z cher $ */

#ifndef __REUSE_PRINTF_H__
#define __REUSE_PRINTF_H__

/* Copyright (C) 1995-2004 Alexander Chernov <cher@ispras.ru> */
/* Created: <95/12/04 20:00:41 cher> */

/*
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 */

/* This module defines printf-like functions for formatted output.
 * These functions support all standard printf format specifiers and
 * add new format specifiers for printing data types of Reuse library,
 * such as Sets, Idents and Boolean values. User can define his own
 * specifiers and use them in format line.
 * Limitations of implementation:
 * 1) all characters between a '%' sign and
 * one of format letters are considered as format specifier and are passed
 * to corresponding routine (to `printf' in case of standard format letter
 * or to user function in case of user-defined format letters). Therefore
 * format letters cannot be used in format string. There are some particular
 * letters that are used in standard format convertions as format modifiers.
 * These letters cannot be used in user-defined formats. These letters are:
 * `h', `l', `L', 'w';
 * 2) positional format specifiers (that begin for `digits$') are not
 * supported;
 * 3) asterisk conversions (that use `*' sign to indicate that width or
 * precision is placed in parameters) are supported;
 * 4) %n specification will always return 0;
 *
 * Auxiluary format conversions:
 *       %[spec.]I  -- prints an identifier of type tIdent.
 *                     spec. is a witdth and/or precision specification
 *                     valid for %s format specification.
 *                     NOTE: asterisks are not supported here.
 *
 */

#include <reuse/Defs.h>
#include <reuse/ReuseDefs.h>
#include <reuse/GenOut.h>

#include <stdarg.h>

typedef int (*tUserPrintfFunc) (tGenOut *, char *, int, va_list);
/* User formatted output funtion type. When it is called the followong
 * values are passed as parameters:
 * 
 * -- pointer to Generic Output structure as a destination of output.
 * User have to use functions from `GenOut' module to write to it.
 * 
 * -- format specifier with leading '%' sign and with trailing format
 * specifier letter;
 * 
 * -- length of the format string
 * 
 * -- va_list of auxiluary arguments in stack.
 * 
 * -- function must return how many arguments it fetched from the stack.
 * Function must return their total size in bytes.
 */

#if defined __cplusplus
extern "C" {
#endif /* __cplusplus */

  R_DLL void GPrintf  (tGenOut *, char *, ...);
  R_DLL void VGPrintf (tGenOut *, char *, va_list);

  R_DLL Bool Printf_Register (int letter, tUserPrintfFunc func);
/* register user-defined format conversion.
 * -- letter is a format letter;
 * -- func is user function to be called when format letter is occured
 * in string;
 * 
 * function returns True if function is successfully added and
 * False if function cannot be added (specified letter is already busy or
 * is not a letter at all.
 */

  R_DLL void prfInitModule(void);
/* initializes module. This routine must be called before use
 * functions from the module.
 */

  R_DLL void prfCloseModule(void);
/* closes module. Frees memory allocated for internal data. */

  R_DLL void prfStatistics(FILE *);
/* prints some statistics.
 * NOTE: you may call this function both after closing of the module
 * and during its work
 * Currently this function does nothing.
 */

#if defined __cplusplus
}
#endif /* __cplusplus */

#endif /* __REUSE_PRINTF_H__ */

/**
 * Local variables:
 *  compile-command: "make -C ../.."
 *  c-font-lock-extra-types: ("tIdent" "FILE" "\\sw+_t" "tGenOut" "va_list" "Bool" "tUserPrintfFunc" "R_DLL")
 * End:
 */
