/* $Id: dyn_array.h 305 2002-06-15 22:18:12Z cher $ */

#ifndef __REUSE_DYN_ARRAY_H__
#define __REUSE_DYN_ARRAY_H__

/* Copyright (C) 1999-2002 Alexander Chernov <cher@ispras.ru> */
/* Copyright (C) 1997 SDL Research Lab, Moscow */

/*
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 */

/**
 * FILE:        dyn_array.h
 * AUTHOR:      dmjr
 * SYSTEM:      Template Classes
 * DESCRIPTION: The dynamic array class library
 */

/* -----------------------Include-files----------------------- */


#include <stdlib.h>

#include <reuse/ReuseDefs.h>


/* ---------------------Macro-definitions--------------------- */


/**
 * NAME:      DA_DECLARE
 * PURPOSE:   declare a dynamic array
 * SYNOPSIS:  arr_type -- array type
 *            eltType -- element type
 */

#define DA_DECLARE(arr_type,eltType) \
  typedef struct arr_type { \
    eltType *v; \
    size_t size, alloc; \
  } arr_type;

/**
 * NAME:      DA_INIT
 * PURPOSE:   initialize a dynamic array
 * SYNOPSIS:  array -- array being initialized
 * RETURNS:   void
 */

#define DA_INIT(array) \
  { (array).v = NULL; (array).size = (array).alloc = 0; }

/**
 * NAME:      DA_FREE
 * PURPOSE:   finalize a dynamic array
 * SYNOPSIS:  array -- array being finalized
 * RETURNS:   void
 * NOTE:      frees all memory used by an array
 */

#define DA_FREE(array) \
  free((array).v);

/**
 * NAME:      DA_FREEZE
 * PURPOSE:   release the most possible memory amount used by array
 * SYNOPSIS:  array -- the array
 * RETURNS:   void
 */

#define DA_FREEZE(array,eltType) \
  da_Freeze((tpDynamicArray)&(array),sizeof(*(array).v));

/**
 * NAME:      DA_COPY
 * PURPOSE:   copy one array to another
 * SYNOPSIS:  dst -- the source array
 *            src -- the destanation array
 * RETURNS:   void
 */

#define DA_COPY(dst,src) \
  da_Copy((tpDynamicArray)&(dst),(tpDynamicArray)&(src),sizeof(*(dst).v));

/**
 * NAME:      DA_TRUNC
 * PURPOSE:   truncate an array to zero length
 * SYNOPSIS:  array -- array being truncated
 * RETURNS:   void
 */

#define DA_TRUNC(array) \
  (array).size = 0;

/**
 * NAME:      DA_ADD
 * PURPOSE:   add an element to the end of an array
 * SYNOPSIS:  array -- the array
 *            eltType -- element type
 * RETURNS:   eltType * -- pointer to added element
 */

#define DA_ADD(array,eltType) \
  (eltType *)da_Add((tpDynamicArray)&(array),sizeof(eltType))

/**
 * NAME:      DA_EXTEND
 * PURPOSE:   extend an array to given size
 * SYNOPSIS:  array -- the array
 *            newSize -- the new size
 * RETURNS:   void
 * NOTE:      if the array already contains more than size elements
 *                macro does nothing
 */

#define DA_EXTEND(array,newSize) \
  da_Extend((tpDynamicArray)&(array),newSize,sizeof(*(array).v))

/**
 * NAME:      DA_REMOVE
 * PURPOSE:   remove an element of an array
 * SYNOPSIS:  array -- the array
 *            index -- index of element being removed
 * RETURNS:   void
 * NOTE:      doesn't preserve order of array elements
 */

#define DA_REMOVE(array,index) \
  (array).v[index] = (array).v[--(array).size];

/**
 * NAME:      DA_ORD_REMOVE
 * PURPOSE:   remove an element of an array
 * SYNOPSIS:  array -- the array
 *            index -- index of element being removed
 * RETURNS:   void
 * NOTE:      preserves order of array elements
 */

#define DA_ORD_REMOVE(array,index) \
  memcpy((array).v+(index),(array).v+(index)+1, \
         (--(array).size-(index))*sizeof(*(array).v));

/**
 * NAME:      DA_APPEND_ARRAY
 * PURPOSE:   append one array to another
 * SYNOPSIS:  dst -- destanation array
 *            src -- array being appended
 * RETURNS:   void
 */

#define DA_APPEND_ARRAY(dst,src) \
  da_AppendMem((tpDynamicArray)&(dst),(src).v, (src).size, \
                 sizeof(*(dst).v));

/**
 * NAME:      DA_APPEND_MEM
 * PURPOSE:   append memory vector to an array
 * SYNOPSIS:  dst -- destanation array
 *            vec -- pointer to vector
 *            vecSize -- size of the vector
 * RETURNS:   void
 */

#define DA_APPEND_MEM(dst,vec,vecSize) \
  da_AppendMem((tpDynamicArray)&(dst),(vec),(vecSize), \
               sizeof(*(dst).v));

/* ----------------------Type-definitions--------------------- */


DA_DECLARE (tDynamicArray, char);
typedef tDynamicArray * tpDynamicArray;

DA_DECLARE (tIntArray, int);
typedef tIntArray * tpIntArray;

/* --------------------Function-prototypes-------------------- */


R_DLL void *da_Add(tpDynamicArray array, size_t eltSize);
R_DLL void da_Copy(tpDynamicArray dst, tpDynamicArray src,
                   size_t eltSize);
R_DLL void da_Extend(tpDynamicArray array, size_t newSize,
                     size_t eltSize);
R_DLL void da_Freeze(tpDynamicArray array, size_t eltSize);
R_DLL void da_Concat(tpDynamicArray dst, tpDynamicArray src,
                     size_t eltSize);
R_DLL void da_AppendMem(tpDynamicArray dst, void *vec, size_t vecSize,
                        size_t eltSize);

#endif /* __REUSE_DYN_ARRAY_H__ */

/*
 * Local variables:
 *  compile-command: "make -C .."
 *  c-font-lock-extra-types: ("\\sw+_t" "FILE" "tpDynamicArray" "R_DLL")
 * End:
 */
