/* $Id: lists.c 354 2004-07-17 06:28:14Z cher $ */

/* Copyright (C) 1998-2004 Alexander Chernov <cher@ispras.ru> */
/* Created: <1998-03-17 18:06:55 cher> */

/*
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 */

/**
 * FILE:    utils/lists.c
 * PURPOSE: generic lists support
 */

#define __REUSE__ 1

/* internal include directives */
#include "lconfig.h"
#include "reuse/config.h"
#include "embed_version.h"

#include <reuse/lists.h>
#include <reuse/logger.h>

/**
 * NAME:    list_Find
 * PURPOSE: find n'th element in the list
 * ARGS:    l - list
 *          n - number of the element 
 * RETURN:  pointer to the list entry
 * NOTE:    if n >= 0, elements are counted from the beginning, ie
 *            0 - the first element in the list
 *            if n is out of the list length, the last element is returned
 *          if n <  0, elements are counted from the end,
 *            -1 - the last element in the list
 *            if n is out of the list length, the first element is returned
 *          
 */
  void *
list_Find(l, n)
     tGenList *l;
     int       n;
{
  tGenListEntry *p;

  ASSERT(l);
  ASSERT(!!l->first == !!l->last);

  if (n < 0)
    {
      for (p = l->last, n++; p && n; n++, p = p->prev) {}
    }
  else
    {
      for (p = l->first; p && n; n--, p = p->next) {}
    }
  return (void*) p;
}

/**
 * NAME:    list_Insert
 * PURPOSE: l - list
 *          n - number in the list to insert to
 *          e - new list item
 * NOTE:    if n <  0, items are counted from the last item
 *            -1           - insert AFTER the last item
 *            n is too big - insert AFTER the first item
 *          if n >= 0, items are counted from the first item
 *            0            - insert BEFORE the first item
 *            n is too big - insert BEFORE the last item
 */
  void
list_Insert(l, n, e)
     tGenList      *l;
     int            n;
     tGenListEntry *e;
{
  tGenListEntry *b;
  tGenListEntry *a;

  ASSERT(l);
  ASSERT(e);
  ASSERT(!!l->first == !!l->last);

  if (n < 0)
    {
      for (a = l->last, n++; a && n; n++, a = a->prev) {}
      if (!a) b = l->first;
      else    b = a->next;
    }
  else
    {
      for (b = l->first; b && n; n--, b = b->next) {}
      if (!b) a = l->last;
      else    a = b->prev;
    }

  if (!a && !b)
    {
      /* the only entry */
      ASSERT(!l->first && !l->last);

      l->first = e;
      l->last  = e;
      e->next  = 0;
      e->prev  = 0;
    }
  else if (!a)
    {
      /* the first entry */
      e->next  = l->first;
      e->prev  = 0;
      l->first = e;
      b->prev  = e;
    }
  else if (!b)
    {
      /* the last entry */
      e->next  = 0;
      e->prev  = l->last;
      l->last  = e;
      a->next  = e;
    }
  else
    {
      e->next = b;
      e->prev = a;
      a->next = e;
      b->prev = e;
    }
}

/**
 * NAME:    list_Remove
 * PURPOSE: remove the n'th item of the list
 * ARGS:    l - list
 *          n - item number
 * RETURN:  pointer to the removed element
 * NOTE:    if n <  0, items are counted from the last item
 *            n == -1      - the last item
 *            n is too big - the first item
 *          if n >= 0, items are counted from the first item
 *            n == 0       - the first item
 *            n is too big - the last item
 */
  void*
list_Remove(l, n)
     tGenList      *l;
     int            n;
{
  tGenListEntry *p;

  ASSERT(l);
  ASSERT(!!l->first == !!l->last);

  if (n < 0)
    {
      for (p = l->last, n++; p && n; n++, p = p->prev) {}
    }
  else
    {
      for (p = l->first; p && n; n--, p = p->next) {}
    }

  if (!p) return 0;

  if (!p->prev && !p->next)
    {
      ASSERT(l->first == p);
      ASSERT(l->last  == p);

      l->first = 0;
      l->last  = 0;
    }
  else if (!p->prev)
    {
      ASSERT(l->first == p);
      ASSERT(l->last  != p);
      ASSERT(p->next);

      p->next->prev = 0;
      l->first      = p->next;
    }
  else if (!p->next)
    {
      ASSERT(l->first != p);
      ASSERT(l->last  == p);
      ASSERT(p->prev);

      p->prev->next = 0;
      l->last       = p->prev;
    }
  else
    {
      ASSERT(l->first != p);
      ASSERT(l->last  != p);

      p->prev->next = p->next;
      p->next->prev = p->prev;
    }
  return p;
}

#if REUSE_COMPILE_CVSID - 0 != 0
static char const cvsid[] =
"$Id: lists.c 354 2004-07-17 06:28:14Z cher $";
#endif /* REUSE_COMPILE_CVSID */

/* === Do not add stuff after this line === */

/*
 * Local variables:
 *  compile-command: "make"
 *  c-font-lock-extra-types: ("\\sw+_t" "FILE" "tGenList" "tGenListEntry")
 * End:
 */
