/* $Id: reuse-config-c.c 426 2006-05-07 09:52:28Z cher $ */

/* Copyright (C) 2005,2006 Alexander Chernov <cher@ispras.ru> */

/*
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>

#include "reuse-config-c.h"

static unsigned char *var_prefix;
static unsigned char *var_exec_prefix;
static unsigned char *var_bindir;
static unsigned char *var_datadir;
static unsigned char *var_includedir;
static unsigned char *var_libdir;
static unsigned char *var_libexecdir;
static unsigned char *var_build;
static unsigned char *var_build_version;
static unsigned char *var_libcap_name;

static unsigned char *prog_name;

struct vartable
{
  const unsigned char *name;
  unsigned char **pvalue;
};

static struct vartable vars[] =
{
  { (const unsigned char*) "prefix", &var_prefix },
  { (const unsigned char*) "exec_prefix", &var_exec_prefix },
  { (const unsigned char*) "bindir", &var_bindir },
  { (const unsigned char*) "datadir", &var_datadir },
  { (const unsigned char*) "includedir", &var_includedir },
  { (const unsigned char*) "libdir", &var_libdir },
  { (const unsigned char*) "libexecdir", &var_libexecdir },
  { (const unsigned char*) "build", &var_build },
  { (const unsigned char*) "build_version", &var_build_version },
#if defined REUSE_CONFIG_libcap
  { (const unsigned char*) "libcap_name", &var_libcap_name },
#endif
  { 0, 0 },
};

static void err(const char *format, ...) __attribute__((noreturn));
static void
err(const char *format, ...)
{
  unsigned char buf[1024];
  va_list args;

  va_start(args, format);
  vsnprintf((char*)buf, sizeof(buf), format, args);
  va_end(args);

  fprintf(stderr, "%s: %s\n", prog_name, buf);
  exit(1);
}

static void reuse_out_of_mem(void) __attribute__((noreturn));
static void
reuse_out_of_mem(void)
{
  err("failed to allocate more memory!");
}

static void reuse_null_size(void) __attribute__((noreturn));
static void
reuse_null_size(void)
{
  err("null size allocation requested!");
}

static void *
xmalloc(size_t size)
{
  void *ptr;

  if (size == 0) reuse_null_size();
  ptr = malloc(size);
  if (ptr == NULL) reuse_out_of_mem();
  return ptr;
}

static const unsigned char *
get_var_value(const unsigned char *varname)
{
  int i;
  const unsigned char *valstr = 0;

  // search in global variables
  for (i = 0; vars[i].name; i++) {
    if (!strcmp((char*)vars[i].name, (char*)varname)) break;
  }
  if (!vars[i].name)
    err("variable `%s' is unknown", vars[i].name);
  if (!*vars[i].pvalue)
    err("variable `%s' is not yet defined", vars[i].name);
  return *vars[i].pvalue;
}

static unsigned char *
varsubst_heap(unsigned char *in_str, int free_flag)
{
  unsigned char *out_str = 0, *p1, *p2;
  unsigned char *var_name = 0;
  size_t var_name_size = 0, in_str_len = 0, var_value_len = 0;
  const unsigned char *var_value = 0;

  //fprintf(stderr, ">>%s\n", in_str);

  in_str_len = strlen((char*)in_str);
  while (1) {
    // find the first variable use
    p1 = in_str;
    for (p1 = in_str; *p1; p1++) {
      if (*p1 == '$' && p1[1] == '{') break;
    }
    if (!*p1) break;
    p2 = p1 + 2;
    while (*p2 && *p2 != '}') p2++;
    if (!*p2 || p2 == p1 + 2) {
      err("varsubst_heap: invalid variable name in %s", in_str);
      if (free_flag) free(in_str);
      free(var_name);
      return 0;
    }
    if (p2 - p1 > var_name_size) {
      var_name_size = p2 - p1;
      free(var_name);
      var_name = xmalloc(var_name_size);
    }
    memcpy(var_name, p1 + 2, p2 - p1 - 2);
    var_name[p2 - p1 - 2] = 0;
    var_value = get_var_value(var_name);
    if (!var_value) {
      if (free_flag) free(in_str);
      free(var_name);
      return 0;
    }
    var_value_len = strlen((char*)var_value);
    out_str = xmalloc(in_str_len + var_value_len + 1);
    memcpy(out_str, in_str, p1 - in_str);
    memcpy(out_str + (p1 - in_str), var_value, var_value_len);
    strcpy((char*) out_str + (p1 - in_str) + var_value_len, (char*)p2 + 1);

    if (free_flag) free(in_str);
    in_str = out_str;
    in_str_len = strlen((char*)in_str);
    free_flag = 1;
  }
  free(var_name);

  //fprintf(stderr, ">>%s\n", in_str);

  // find whether there are dollar substitutions
  for (p1 = in_str; *p1; p1++) {
    if (*p1 == '$' && p1[1] == '$') break;
  }
  if (!*p1) return in_str;

  out_str = xmalloc(in_str_len + 1);
  for (p1 = in_str, p2 = out_str; *p1; p1++, p2++) {
    *p2 = *p1;
    if (*p1 == '$' && p1[1] == '$') p1++;
  }
  if (free_flag) free(in_str);
  return out_str;
}

static void usage(void) __attribute__((noreturn));
static void
usage(void)
{
  fputs("Usage: reuse-config [OPTION]...\n"
        "\n"
        "General options:\n"
        "  --help               print this message\n"
        "  --version            print the version of reuse associated with this script\n"
        "  --build              print the canonic configuration name\n"
        "  --prefix             print the installation prefix\n"
        "  --cflags             print C the compilation flags for reuse\n"
        "  --ldflags            print C the linking flags for reuse\n"
        "  --libs               print C the libraries for reuse\n",
        stdout);

  exit(0);
}

static int need_cflags;
static int need_ldflags;
static int need_libs;

int
main(int argc, char **argv)
{
  int i;
  unsigned char tmpbuf[8192];
  unsigned char outbuf[8192] = { 0 };
  unsigned char *sps;
  
  prog_name = (unsigned char*) argv[0];

  var_prefix = varsubst_heap((unsigned char*)REUSE_CONFIG_prefix, 0);
  var_exec_prefix = varsubst_heap((unsigned char*)REUSE_CONFIG_exec_prefix, 0);
  var_bindir = varsubst_heap((unsigned char*)REUSE_CONFIG_bindir, 0);
  var_datadir = varsubst_heap((unsigned char*)REUSE_CONFIG_datadir, 0);
  var_includedir = varsubst_heap((unsigned char*)REUSE_CONFIG_includedir, 0);
  var_libdir = varsubst_heap((unsigned char*)REUSE_CONFIG_libdir, 0);
  var_libexecdir = varsubst_heap((unsigned char*)REUSE_CONFIG_libexecdir, 0);
  var_build = varsubst_heap((unsigned char*)REUSE_CONFIG_build, 0);
  var_build_version = varsubst_heap((unsigned char*)REUSE_CONFIG_build_version, 0);
#if defined REUSE_CONFIG_libcap
  var_libcap_name = varsubst_heap((unsigned char*)REUSE_CONFIG_libcap, 0);
#endif

  if (argc <= 1) usage();

  if (!strcmp(argv[1], "--help")) usage();
  if (!strcmp(argv[1], "--version")) {
    printf("%s\n", var_build_version);
    exit(0);
  }
  if (!strcmp(argv[1], "--build")) {
    printf("%s\n", var_build);
    exit(0);
  }
  if (!strcmp(argv[1], "--prefix")) {
    printf("%s\n", var_prefix);
    exit(0);
  }
  if (!strcmp(argv[1], "--vars")) {
    printf("prefix=%s\n"
           "exec_prefix=%s\n"
           "bindir=%s\n"
           "datadir=%s\n"
           "includedir=%s\n"
           "libdir=%s\n"
           "libexecdir=%s\n"
           "build=%s\n"
           "build_version=%s\n",
           var_prefix, var_exec_prefix, var_bindir, var_datadir,
           var_includedir, var_libdir, var_libexecdir, var_build,
           var_build_version);
    exit(0);
  }

  for (i = 1; i < argc; i++) {
    if (!strcmp("--cflags", argv[i])) {
      need_cflags = 1;
    } else if (!strcmp("--ldflags", argv[i])) {
      need_ldflags = 1;
    } else if (!strcmp("--libs", argv[i])) {
      need_libs = 1;
    } else {
      err("invalid option `%s'", argv[i]);
    }
  }

  if (need_cflags) {
    sps = (unsigned char*) "";
    if (outbuf[0]) sps = (unsigned char*) " ";
    snprintf((char*)tmpbuf, sizeof(tmpbuf), "%s%s-I%s -I%s/reuse/include", outbuf, sps,
             var_includedir, var_libdir);
    strcpy((char*) outbuf, (char*) tmpbuf);
  }

  if (need_ldflags) {
    sps = (unsigned char*) "";
    if (outbuf[0]) sps = (unsigned char*) " ";
    snprintf((char*)tmpbuf, sizeof(tmpbuf), "%s%s-L%s", outbuf, sps, var_libdir);
    strcpy((char*)outbuf, (char*)tmpbuf);
#if !defined REUSE_CONFIG_disable_rpath || !REUSE_CONFIG_disable_rpath
    snprintf((char*)tmpbuf, sizeof(tmpbuf), "%s -Wl,-rpath,%s", outbuf, var_libdir);
    strcpy((char*)outbuf, (char*)tmpbuf);
#endif
  }

  if (need_libs) {
    sps = (unsigned char*)"";
    if (outbuf[0]) sps = (unsigned char*)" ";
    if (var_libcap_name && *var_libcap_name) {
      snprintf((char*)tmpbuf, sizeof(tmpbuf), "%s%s-lreuse %s -lm",
               outbuf, sps, var_libcap_name);
    } else {
      snprintf((char*)tmpbuf, sizeof(tmpbuf), "%s%s-lreuse -lm", outbuf, sps);
    }
    strcpy((char*)outbuf, (char*)tmpbuf);
  }

  if (outbuf[0]) printf("%s\n", outbuf);

  return 0;
}

/*
 * Local variables:
 *  compile-command: "make"
 * End:
 */
